# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

import collections
import struct
import threading
from itertools import islice

from zapf.simulator.util import NumProxy

# The Memory is a thread-local singleton.  It sets itself as `mem` on this
# object on construction.
local = threading.local()

# ruff: noqa: N801


class Memory:
    # Allocates addresses for variables.

    def __init__(self):
        # %M* is at 0x10000
        # %I* is at 0x20000
        # %Q* is at 0x30000
        # dyn is at 0x40000 up
        self.dyn_addr = 0x40000
        self.allocations = []
        self.allocations_sorted = False
        self.allocs_by_size = {}
        local.mem = self

    def sort_allocations(self):
        self.allocs_by_size.clear()
        self.allocations.sort(key=lambda a: a[:2])
        # Allocations are sorted by size first, then by address - this lets us
        # find the "innermost" allocation for a given byte range.  But in order
        # to speed up this search, we cache the index where to start the search
        # for any given size.
        last = 0
        for (i, (osize, _, _)) in enumerate(self.allocations):
            if osize > last:
                for nsize in range(last + 1, osize + 1):
                    self.allocs_by_size[nsize] = i
                last = osize
        self.allocations_sorted = True

    def new(self, size):
        addr = self.dyn_addr
        self.dyn_addr += size
        return addr

    def map(self, addr, obj):
        """Map a new allocation."""
        self.allocations.append((obj.sizeof(), addr, obj))
        self.allocations_sorted = False

    def get(self, addr, size):
        """Get object.

        Return the object that addr belongs to, and the offset in it,
        if there are at least size bytes left.
        """
        if not self.allocations_sorted:
            self.sort_allocations()
        # Find the smallest allocation that encompasses the whole given range.
        # If the size is not in "alloc_size_start", the requested size is
        # bigger than any single object, which is not allowed.
        if size in self.allocs_by_size:
            allocs = islice(self.allocations, self.allocs_by_size[size], None)
            for (osize, oaddr, obj) in allocs:
                if oaddr <= addr and addr + size <= oaddr + osize:
                    return (obj, addr - oaddr)
        raise RuntimeError('no value or addressing across value boundary')

    def read(self, addr, size):
        """Read memory."""
        (obj, offset) = self.get(addr, size)
        return obj.mem_read()[offset : offset + size]

    def write(self, addr, data):
        """Write memory."""
        (obj, offset) = self.get(addr, len(data))
        obj.mem_write(offset, data)


class Value:
    """Represents a place in memory for a variable."""

    @classmethod
    def alloc(cls, value, at=None):
        """Allocates an address for the value (if not given)."""
        if at is None:
            at = local.mem.new(cls.sizeof())
        val = cls(value, at)
        local.mem.map(at, val)
        return val

    def alloc_self(self, at=None):
        """Shortcut for self.alloc(self)."""
        return self.alloc(self, at)

    @classmethod
    def default(cls, at=None):
        """Allocates a value with a default value."""
        return cls.alloc(cls.DEFAULT, at)

    @classmethod
    def unwrap(cls, value):
        """Retrieve the inner value from an rvalue assigned to this lvalue."""
        if isinstance(value, cls):
            return value.value
        return value

    @classmethod
    def sizeof(cls):
        """Return the size of the value in memory."""
        raise NotImplementedError

    def __init__(self, value, addr):
        self.addr = addr
        self.assign(value)

    def __repr__(self):
        return repr(self.value)

    def assign(self, value):
        self.value = self.__class__.unwrap(value)

    def mem_read(self):
        raise NotImplementedError

    def mem_write(self, offset, data):
        raise NotImplementedError


class Integral(NumProxy, Value):
    DEFAULT = 0
    WIDTH = 0
    SIGNED = False
    MEMFMT = ''

    @classmethod
    def sizeof(cls):
        return cls.WIDTH // 8

    @classmethod
    def unwrap(cls, value):
        if isinstance(value, Integral):
            value = value.value
        limit = 1 << cls.WIDTH
        value %= limit
        if cls.SIGNED and value >= limit // 2:
            value -= limit
        # if cls.SIGNED:
        #     limit = 1 << (cls.WIDTH - 1)
        #     if not -limit <= value < limit:
        #         raise RuntimeError('out of range assignment to %s: %s' % (
        #             cls.__name__, value))
        # else:
        #     if value < 0 or value >= 1 << cls.WIDTH:
        #         raise RuntimeError('out of range assignment to %s: %s' % (
        #             cls.__name__, value))
        return value

    def mem_read(self):
        return struct.pack(self.MEMFMT, self.value)

    def mem_write(self, offset, data):
        if offset != 0:
            raise RuntimeError(f'partial {self.__class__.__name__} write '
                               f'(offset {offset}, datalen {len(data)})')
        self.value, = struct.unpack(self.MEMFMT, data)

    def __getitem__(self, i):
        # Bit access: a[[i]]
        return (self.value >> i[0]) & 1

    def __setitem__(self, i, val):
        mask = ~(1 << i[0])
        self.value = (self.value & mask) | ((val & 1) << i[0])


class byte(Integral):
    WIDTH = 8
    SIGNED = False
    MEMFMT = 'B'


class word(Integral):
    WIDTH = 16
    SIGNED = False
    MEMFMT = 'H'


class dword(Integral):
    WIDTH = 32
    SIGNED = False
    MEMFMT = 'I'


class lword(Integral):
    WIDTH = 64
    SIGNED = False
    MEMFMT = 'Q'


class boolean(Integral):
    WIDTH = 1
    SIGNED = False
    MEMFMT = 'B'

    @classmethod
    def sizeof(cls):
        return 1


class real(NumProxy, Value):
    DEFAULT = 0.0

    @classmethod
    def sizeof(cls):
        return 4

    def mem_read(self):
        return struct.pack('f', self.value)

    def mem_write(self, offset, data):
        if offset != 0:
            raise RuntimeError(f'partial {self.__class__.__name__} write '
                               f'(offset {offset}, datalen {len(data)})')
        self.value, = struct.unpack('f', data)


class lreal(NumProxy, Value):
    DEFAULT = 0.0

    @classmethod
    def sizeof(cls):
        return 8

    def mem_read(self):
        return struct.pack('d', self.value)

    def mem_write(self, offset, data):
        if offset != 0:
            raise RuntimeError(f'partial {self.__class__.__name__} write '
                               f'(offset {offset}, datalen {len(data)})')
        self.value, = struct.unpack('d', data)


class anystring(Value):
    SLEN = 0
    DEFAULT = ''

    @classmethod
    def sizeof(cls):
        return cls.SLEN

    @classmethod
    def unwrap(cls, value):
        if isinstance(value, anystring):
            value = value.value
        if len(value.encode()) >= cls.SLEN:
            raise RuntimeError(f'string too long ({cls.SLEN-1} bytes max)')
        return value

    def mem_read(self):
        return self.value.encode() + b'\0' * (self.SLEN - len(self.value.encode()))

    def mem_write(self, offset, data):
        if offset != 0:
            raise RuntimeError(f'partial {self.__class__.__name__} write '
                               f'(offset {offset}, datalen {len(data)})')
        self.value = data.split('\0', 1)[0]

    def __len__(self):
        return self.value.__len__()


def string(slen):
    # string(X) adds a trailing null byte, so size is X+1
    return type(f'string_{slen}', (anystring,), {'SLEN': slen + 1})


class anyarray(Value):
    LENGTH = 0
    IMIN = 0
    INNER = None
    DEFAULT = []  # noqa: RUF012

    @classmethod
    def sizeof(cls):
        return cls.LENGTH * cls.INNER.sizeof()

    def __init__(self, value, addr):
        self.addr = addr
        self.value = []
        value = self.__class__.unwrap(value)
        step = self.INNER.sizeof()
        if len(value) > self.LENGTH:
            raise RuntimeError(f'too many values in array assignment '
                               f'(max {self.LENGTH})')
        for i in range(self.LENGTH):
            self.value.append(self.INNER.alloc(
                value[i] if i < len(value) else self.INNER.DEFAULT,
                at=addr + i * step))

    def assign(self, _value):
        raise RuntimeError('array assign')

    def __getitem__(self, i):
        if not self.IMIN <= i < self.IMIN + self.LENGTH:
            raise RuntimeError(f'array access out of range (index {i}, '
                               f'min {self.IMIN}, length {self.LENGTH})')
        return self.value[i - self.IMIN]

    def __setitem__(self, i, val):
        if not self.IMIN <= i < self.IMIN + self.LENGTH:
            raise RuntimeError(f'array access out of range (index {i}, '
                               f'min {self.IMIN}, length {self.LENGTH})')
        self.value[i - self.IMIN].assign(val)

    def __len__(self):
        return self.LENGTH

    def __repr__(self):
        return f'<{", ".join(map(repr, self.value))}>'

    def mem_read(self):
        return b''.join(v.mem_read() for v in self.value)

    def mem_write(self, offset, data):
        step = self.INNER.sizeof()
        if offset % step != 0:
            raise RuntimeError('partial array element write')
        if len(data) % step != 0:
            raise RuntimeError('partial array element write')
        i = offset // step
        while data:
            d, data = data[:step], data[step:]
            self.value[i].mem_write(0, d)
            i += 1


def array(innertype, imin, imax):
    length = imax - imin + 1
    return type(f'array_{length}', (anyarray,), {'LENGTH': length,
                                                 'IMIN': imin,
                                                 'INNER': innertype})


class Var:
    """Represents a field in a struct.

    This includes the "variable struct" automatically generated for a function
    decorated with @program.
    """

    def __init__(self, dtype, default=None, *, at=None):
        self.dtype = dtype
        self.default = default if default is not None else dtype.DEFAULT
        self.at = None
        if at is not None:
            if at.startswith('%MB'):
                self.at = int(at[3:]) + 0x10000
            elif at.startswith('%IB'):
                self.at = int(at[3:]) + 0x20000
            elif at.startswith('%QB'):
                self.at = int(at[3:]) + 0x30000
            else:
                raise RuntimeError(f'addr spec {at!r} not supported')

    def __get__(self, obj, obj_class):
        if obj is None:
            return self
        return obj.__dict__[self]

    def __set__(self, obj, value):
        if obj is None:
            return
        obj.__dict__[self].assign(self.dtype.unwrap(value))


class StructMeta(type):
    @classmethod
    def __prepare__(mcs, _name, _bases):  # noqa: N804
        return collections.OrderedDict()

    def __init__(cls, _name, _bases, attrs):
        cls.VARS = []
        cls.OFFSET = {}
        size = 0
        for (varname, var) in attrs.items():
            if isinstance(var, Var):
                cls.VARS.append((varname, var))
                cls.OFFSET[varname] = size
                # XXX alignment!
                size += var.dtype.sizeof()
        cls.SIZE = size


class Struct(Value, metaclass=StructMeta):
    DEFAULT = Ellipsis

    @classmethod
    def sizeof(cls):
        return cls.SIZE

    def __init__(self, value=None, addr=None, **pvars):
        self.addr = addr
        for (name, var) in self.VARS:
            self.__dict__[var] = var.dtype.alloc(
                getattr(value, name, pvars.pop(name, var.default)),
                at=var.at or (addr + self.OFFSET[name]
                              if addr is not None else None))
        if pvars:
            raise RuntimeError(f'unknown variable(s) in struct: {pvars}')

    def assign(self, value):
        if value is Ellipsis:
            # XXX set defaults necessary?
            # keep defaults
            return
        if isinstance(value, self.__class__):
            # XXX reassign individual values!
            raise RuntimeError('struct reassign')  # noqa: TRY004
        raise RuntimeError('trying to assign struct of wrong kind')

    def __repr__(self):
        items = []
        for (name, var) in self.VARS:
            items.append((name, self.__dict__[var]))
        return self.__class__.__name__ + ' { ' + \
            ', '.join(f'{x[0]} => {x[1]!r}' for x in items) + ' }'

    def mem_read(self):
        return b''.join(self.__dict__[var].mem_read()
                        for (_, var) in self.VARS)

    def mem_write(self, offset, data):
        for (name, var) in self.VARS:
            var_skip = offset - self.OFFSET[name]
            var_size = var.dtype.sizeof()
            # skip elements before the first offset
            if var_skip >= var_size:
                continue
            write_size = min(var_size - var_skip, len(data))
            d, data = data[:write_size], data[write_size:]
            self.__dict__[var].mem_write(var_skip, d)
            offset += write_size
            if not data:
                break


class Globals(Struct):
    pass


def program(**pvars):
    def deco(func):
        var_struct = type(f'{func.__name__}_vars', (Struct,), pvars)
        instance = var_struct()

        def new_func():
            func(instance)

        new_func.is_program = True
        return new_func
    return deco



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Basic test suite for the devices in the testplc."""

# pylint: disable=protected-access
# ruff: noqa: SLF001

import time

import pytest

from zapf import ApiError, CommError
from zapf.device import typecode_description
from zapf.spec import DevStatus, ParamCMDs, decode_bitfield


def test_readable_device(plc_readable_device):
    """Test value/status."""
    # zapf-device, TYPCODE_MAP entry
    dev, _tcme = plc_readable_device

    state, reason, aux, err_id = dev.read_status()
    decode_bitfield(aux, dev.aux_spec)
    decode_bitfield(err_id, dev.errid_spec)

    assert state in [
        DevStatus.IDLE,
        DevStatus.DISABLED,
        DevStatus.WARN,
        DevStatus.START,
        DevStatus.BUSY,
        DevStatus.STOP,
        DevStatus.ERROR,
    ]
    assert typecode_description(dev.typecode).startswith(dev.name)
    assert reason in [0, 1, 2, 4, 8]
    assert err_id == 0

    value = dev.read_value_raw()
    limits = dev.value_info.min_value, dev.value_info.max_value
    assert limits[0] <= value <= limits[1]

    if not dev.target_addr:
        with pytest.raises(ApiError):
            dev.read_target_raw()
        with pytest.raises(ApiError):
            dev.change_target_raw(0)

    if dev.tables:
        # name is specific to testplc...
        assert 'table1' in dev.tables
        assert 'table1' in dev.list_tables()
        for tbl in dev.tables.values():
            nrows, ncols = tbl.get_size()
            collist = tbl.list_columns()
            assert ncols == len(collist)
            # test tables are square (10x10)
            for i, c in enumerate(collist):
                tbl.get_column_valueinfo(c)
                tbl.get_cell(i, c)
                tbl.get_cell(nrows - 1 - i, c)
            assert tbl.list_columns()[3] == 'WarnMax'
            with tbl as t:
                assert t._last_row is None
                t.set_cell(3, 'WarnMax', 98.0)
                assert t._last_row is not None
                assert t.get_cell(3, 3) == 98.0
                tbl.set_cell(3, 3, 87.0)
            assert t._last_row is None

            assert tbl.get_cell(3, 'WarnMax') == 87.0
            with pytest.raises(ApiError):
                tbl.get_cell(33, 3)  # row too big
            with pytest.raises(ApiError):
                tbl.get_cell(-3, 3)  # row negative
            with pytest.raises(ApiError):
                tbl.get_cell(3, 33)  # col too big
            with pytest.raises(ApiError):
                tbl.get_cell(3, -3)  # col negative
            with pytest.raises(ApiError):
                tbl.get_cell(3, 'no such column')  # no such column

            with pytest.raises(ApiError):
                tbl.set_cell(33, 3, 3)  # row too big
            with pytest.raises(ApiError):
                tbl.set_cell(-3, 3, 3)  # row negative
            with pytest.raises(ApiError):
                tbl.set_cell(3, 33, 3)  # col too big
            with pytest.raises(ApiError):
                tbl.set_cell(3, -3, 3)  # col negative
            with pytest.raises(ApiError):
                tbl.set_cell(3, 'no such column', 3)  # no such column
            with pytest.raises(ApiError):
                tbl.set_cell(3, 3, 'bad value type')  # bad value type

    for pn in dev.list_params():
        info = dev.get_param_valueinfo(pn)
        assert info.readonly in (True, False)


def test_target_device(plc_target_device):
    """Test target, moving around."""
    # zapf-device, TYPCODE_MAP entry
    dev, _tcme = plc_target_device

    state, _, _, _ = dev.read_status()

    target = dev.read_target_raw()
    limits = dev.value_info.min_value, dev.value_info.max_value
    assert limits[0] <= target <= limits[1]

    # check precondition before starting a movement
    assert state not in [DevStatus.START, DevStatus.BUSY, DevStatus.STOP]
    assert state in [DevStatus.IDLE, DevStatus.WARN]

    value = dev.read_value_raw()
    target = value + 1
    dev.change_target_raw(target)
    assert dev.read_target_raw() == target

    timesout = time.time() + 1
    # now check for end of movement
    while True:
        state, _, _, _ = dev.read_status()
        value = dev.read_value_raw()
        target = dev.read_target_raw()
        assert state in [
            DevStatus.START,
            DevStatus.BUSY,
            DevStatus.IDLE,
            DevStatus.WARN,
        ]
        if value != target:
            assert state == DevStatus.BUSY
        if state != DevStatus.BUSY:
            assert value == target
            return
        assert time.time() < timesout

    # check enum translation
    if dev.value_info.enum_r:
        val = isinstance(dev.read_value(), str)
        tgt = isinstance(dev.read_target(), str)
        assert val == tgt
        dev.change_target(dev.read_target())


def test_vector_devices(plc_vector_device):
    """Test vector value/status/target/moving around."""
    # zapf-device, TYPCODE_MAP entry
    dev, tcme = plc_vector_device

    has_target, num_values = tcme.has_target, tcme.num_values

    state, _, _, _ = dev.read_status()

    value = dev.read_value_raw()
    assert len(value) == num_values

    limits = dev.value_info.min_value, dev.value_info.max_value
    for v in value:
        assert limits[0] <= v <= limits[1]

    if not has_target:
        with pytest.raises(ApiError):
            dev.read_target_raw()
        with pytest.raises(ApiError):
            dev.change_target_raw(value)
        return

    target = dev.read_target_raw()
    for t in target:
        assert limits[0] <= t <= limits[1]
    if target == value:
        assert state in [DevStatus.IDLE, DevStatus.DISABLED, DevStatus.WARN]
    else:
        assert state in [
            DevStatus.DISABLED,
            DevStatus.START,
            DevStatus.BUSY,
            DevStatus.STOP,
            DevStatus.ERROR,
        ]
    dev.change_target_raw([t+1 for t in target])


def test_flat_device(plc_flat_device):
    """Test flat parameter access."""
    # zapf-device, TYPCODE_MAP entry
    dev, tcme = plc_flat_device

    assert tcme.num_params == len(dev.list_params())

    # check param access of all params
    for p in dev.list_params():
        if p in ['Setpoint', 'Offset']:
            continue
        min_value = dev.get_param_valueinfo(p).min_value
        v = dev.get_param(p)
        if min_value > -1e10:
            dev.set_param(p, min_value + 1)
            assert pytest.approx(min_value + 1) == dev.get_param(p)
        else:
            dev.set_param(p, v + 1)
            assert pytest.approx(v + 1) == dev.get_param(p)

        dev.set_param_raw(p, v)
        assert v == dev.get_param_raw(p)[1]

        if min_value > -1e10:  # doesn't work for default limits
            with pytest.raises(ApiError):
                dev.set_param(p, min_value - 1)

    # check that Setpoint is readable, but not writable
    if 'Setpoint' in dev.list_params():
        dev.get_param('Setpoint')
        assert dev.set_param_raw('Setpoint', 12)[0] == ParamCMDs.ERR_RO


def test_paramif(plc_param_device):
    """Test paramif access + funcs."""
    # zapf-device, TYPCODE_MAP entry
    dev, _tcme = plc_param_device

    # testplc implements a few parameters:
    # readable: 1-3, 32-38, 40, 43, 44, 51-53, 55-65, 68-70
    # writeable: 32-37, 40, 51-53, 56, 57, 60, 61, 69, 70
    # executable: 128, 133, 137, 142
    # 128 and 133 almost instantly go from BUSY to DONE, rest is ERR_RETRY

    # check state
    dev.wait_sm_available(cached=True)
    assert dev.param_sm.CMD != 0

    # basic lowlevel tests

    # set a known paramvalue != 0 first
    fmt = dev.get_param_valueinfo(1).fmt
    # initiate reading parameter 1 (should result in value 0)
    dev.set_pctrl(ParamCMDs.DO_READ, 1, 42, fmt)
    assert dev.wait_sm_available(fmt)
    assert dev.param_sm.CMD == ParamCMDs.DONE
    assert dev.param_value not in (1, None)

    # try to read param 127 (should fail with NO_IDX)
    assert dev.wait_sm_available()
    dev.set_pctrl(ParamCMDs.DO_READ, 127)
    assert dev.wait_sm_available()
    assert dev.param_sm.CMD == ParamCMDs.ERR_NO_IDX

    # try to write param 15 (should fail with RO or NO_IDX, depending on magic)
    assert dev.wait_sm_available()
    dev.set_pctrl(ParamCMDs.DO_WRITE, 18, 42,
                                'I' if dev.value_size == 4 else 'Q')
    assert dev.wait_sm_available()
    assert dev.param_sm.CMD in [ParamCMDs.ERR_NO_IDX, ParamCMDs.ERR_RO]

    # try to exec func 142 (should fail with RETRY)
    assert dev.wait_sm_available()
    dev.set_pctrl(ParamCMDs.DO_WRITE, 142)
    assert dev.wait_sm_available()
    assert dev.param_sm.CMD == ParamCMDs.ERR_RETRY

    # try to 'BUSY' param 1 (should succeed with DONE)
    assert dev.wait_sm_available()
    dev.set_pctrl(ParamCMDs.BUSY, 1, 42, fmt)
    assert dev.wait_sm_available(fmt)
    assert dev.param_sm.CMD == ParamCMDs.DONE
    assert dev.param_value not in (42, None)

    # no fmt -> param_value should be None
    assert dev.wait_sm_available()
    assert dev.param_value is None

    # higher level tests
    if 'UserMax' in dev.list_params():
        v = dev.get_param('UserMax')
        dev.set_param('UserMax', v - 1)
        assert dev.get_param('UserMax') == v - 1

        assert dev.get_param_index('UserMax')
        info = dev.get_param_valueinfo('UserMax')
        assert info.basetype == 'float'
        assert info.fmt in ('f', 'd')
        assert info.unit == dev.info['unit']
        if dev.io.indexer.magicstr == '2021_09':
            assert info.description
            assert info.min_value == 0
            assert info.max_value == 100

    if 'microsteps' in dev.list_params():
        info = dev.get_param_valueinfo('microsteps')
        assert 1 in info.enum_r
        assert info.enum_w['4 steps'] == 2
        dev.set_param('microsteps', '4 steps')
        assert dev.get_param_raw('microsteps')[1] == 2

    if dev.list_funcs():
        dev.exec_func(dev.list_funcs()[0])


def test_msgio_device(plc_msgio_device):
    """Test for msgio."""
    # zapf-device, TYPCODE_MAP entry
    dev, _tcme = plc_msgio_device

    assert dev.communicate(b'Uryyb Jbeyq!') == b'Hello World!'
    s = b'"rot13" is soo much fun, how can one livE without it???!'
    assert dev.communicate(dev.communicate(s)) == s

    # request too long -> error
    with pytest.raises(CommError):
        dev.communicate(s + s)

    # check state (should still be Error)
    assert dev.read_status()[0] == DevStatus.ERROR
    # try to reset + check state afterwards
    dev.reset()
    for i in range(2):
        if dev.read_status()[0] != DevStatus.RESET:
            break
        time.sleep(i * 0.001)
    assert dev.read_status()[0] == DevStatus.IDLE

    # now test read_value/change_target
    with pytest.raises(ApiError):
        dev.change_target_raw(b'X')
    with pytest.raises(ApiError):
        dev.read_target_raw()
    with pytest.raises(ApiError):
        dev.read_value_raw()

    assert dev.communicate([b'Uryyb', b' ', b'Jbeyq!']) == [b'Hello World!']
    with pytest.raises(ApiError):
        dev.communicate([s])
    assert dev.communicate([b'YYYYYYYYYYYYYYYYYYYYZZ', b'YY']) == \
        [b'LLLLLLLLLLLLLLLLLLLLMM', b'LL']



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

from zapf import SpecError
from zapf.util import BitFields, UncasedMap

SUPPORTED_MAGICS = ['2014_07', '2015_02', '2021_09']

OFFSET_ADDR = 4
FLOAT32_MAX = 3.402823e38
FLOAT64_MAX = 1.7976931348623158e308
StatusStruct16 = BitFields(STATE=(15, 12), REASON=(11, 8), AUX=(7, 0))
StatusStruct32 = BitFields(STATE=(31, 28), REASON=(27, 24), AUX=(23, 0))
ParamControl = BitFields(available=(15, 15), CMD=(15, 13), IDX=(12, 0))

ReasonMap = [
    '',
    'inhibit',
    'timeout',
    '(inhibit, timeout)',
    'lower limit reached',
    '(inhibit, lower limit)',
    '(timeout, lower limit)',
    '(inhibit, timeout, lower limit)',
    'upper limit reached',
    '(inhibit, upper limit)',
    '(timeout, upper limit)',
    '(inhibit, timeout, upper limit)',
    'both limits reached',
    '(inhibit, both limits)',
    '(timeout, both limits)',
    '(inhibit, timeout, both limits)',
]

ReasonSpec = [[0, 4, '', {i: ReasonMap[i] for i in range(16)}]]


ParamCMDs = UncasedMap(
    INIT=0,
    DO_READ=1,
    DO_WRITE=2,
    BUSY=3,
    DONE=4,
    ERR_NO_IDX=5,
    ERR_RO=6,
    ERR_RETRY=7,
)

DevStatus = UncasedMap(
    RESET=0,
    IDLE=1,
    DISABLED=2,
    WARN=3,
    START=5,
    BUSY=6,
    STOP=7,
    ERROR=8,
    DIAGNOSTIC_ERROR=13,
)

StatusSpec = [[0, 4, '', {i: DevStatus[i]
                          for i in (0, 1, 2, 3, 5, 6, 7, 8, 13)}]]

ErridSpec = [[0, 16, 'errid', {0: ''}]]

FMT_TO_BASETYPE = {
    'h': ('int', 16),
    'H': ('uint', 16),
    'i': ('int', 32),
    'I': ('uint', 32),
    'q': ('int', 64),
    'Q': ('uint', 64),
    'e': ('float', 16),
    'f': ('float', 32),
    'd': ('float', 64),
}

# map basetype to fmt, min_value, max_value
BASETYPE_TO_FMT = {
    ('int', 16):   ('h', -32768, 32767),
    ('uint', 16):  ('H', 0, 65535),
    ('int', 32):   ('i', -2147483648, 2147483647),
    ('uint', 32):  ('I', 0, 4294967295),
    ('int', 64):   ('q', -9223372036854775808, 9223372036854775807),
    ('uint', 64):  ('Q', 0, 18446744073709551615),
    ('float', 16): ('e', -65504, 65504),
    ('float', 32): ('f', -FLOAT32_MAX, FLOAT32_MAX),
    ('float', 64): ('d', -FLOAT64_MAX, FLOAT64_MAX),
    ('enum', 16):  ('H', 0, 65535),
    ('enum', 32):  ('I', 0, 65535),  # enum values are always only 16 bit
    ('enum', 64):  ('Q', 0, 65535),  # independent of their field width
}

# possible flags:
READONLY = 'readonly'
LOWLEVEL = 'lowlevel'


def decode_bitfield(value, spec=None):
    res = []
    if spec is None:
        return value
    for entry in spec:
        if len(entry) == 2:
            lsb, key = entry
            if value & (1 << lsb):
                res.append(key)
        elif len(entry) == 4:
            lsb, width, bfname, enums = entry
            mask = (1 << width) - 1
            bfvalue = (value >> lsb) & mask
            bfvalstr = enums.get(bfvalue, str(bfvalue))
            if bfvalstr:
                if bfname:
                    res.append(f'{bfname}={bfvalstr}')
                else:
                    res.append(bfvalstr)
        else:
            raise SpecError('invalid flag/bitfield specification')
    return ', '.join(res)



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Contains all necessary constants and mappings from the spec for 2012.09 only."""
from __future__ import annotations

import struct
from typing import ClassVar, NamedTuple

from zapf import ApiError, DescError, SpecError
from zapf.spec import BASETYPE_TO_FMT, FLOAT32_MAX, FMT_TO_BASETYPE, LOWLEVEL, READONLY

DEBUG_Descriptor = 32767

DEVICE_FLAGS = {  # bit number to name
    15: LOWLEVEL,
    14: READONLY,
}
for _ in range(14):
    DEVICE_FLAGS[_] = f'unknown flag {_}'

DESCRIPTOR_CLASS_PLC = 1
DESCRIPTOR_CLASS_DEVICE = 2
DESCRIPTOR_CLASS_STRING = 3
DESCRIPTOR_CLASS_ENUM = 4
DESCRIPTOR_CLASS_BF = 5
DESCRIPTOR_CLASS_PAR = 6
DESCRIPTOR_CLASS_FUNC = 6
DESCRIPTOR_CLASS_TABLE = 6
DESCRIPTOR_CLASS_DEBUG = 0xF

TABLE_ACCESS_PCTL   = 'paramctl'
TABLE_ACCESS_MEM_1D = 'memory-mapped:1d'
TABLE_ACCESS_MEM_2D = 'memory-mapped:2d'


class DescriptorMeta(type):
    def __new__(mcs, name, base, attrs):
        newtype = type.__new__(mcs, name, base, attrs)
        newtype.FMT = struct.Struct('<' + ''.join(f[0] for f in
                                                  newtype.FIELDS[:-1]))
        newtype.FMT_SIZE = newtype.FMT.size
        return newtype


class Descriptor(metaclass=DescriptorMeta):
    """Descriptor class, handles all bytes after the descriptor type."""

    # FIELDS: fmt, fieldname, resolve 'c'hain, 'd'escriptor or 's'tring,
    # descr_classes for resolve (descriptor type field is omitted here)
    FIELDS: ClassVar = []

    def __init__(self, descriptor_id, descriptor_type, data):
        self.descriptor_id = descriptor_id
        self.descriptor_type = descriptor_type
        self.is_resolved = False

        fmt_size = self.FMT_SIZE
        if len(data) < fmt_size + 1:  # trailing string is zero
            raise DescError(self, f'not enough data for decoding (need '
                            f'{fmt_size+1} bytes, got {len(data)})')
        if data[-1] != 0:
            raise DescError(self, 'data is not zero padded (last '
                            f'byte: {data[-1]} != 0)')

        # calculate potential padding (fields added later, increasing the
        # stringoffset of the descriptor type)
        padding_len = self.descriptor_type & 0xFF - fmt_size - 2
        if padding_len < 0:
            raise DescError(self, 'descriptor type does not have enough '
                            'data fields (padding would be < 0)')
        self.padding = data[fmt_size:fmt_size + padding_len]
        string = data[fmt_size + padding_len:]
        string = string[:string.index(0)].decode('utf-8', 'backslashreplace')
        setattr(self, self.FIELDS[-1][1], string)

        unpacked_data = self.FMT.unpack_from(data)
        for field, value in zip(self.FIELDS, unpacked_data):
            setattr(self, field[1], value)
        self.decode()

    def __repr__(self):
        return (f'{self.__class__.__name__}/{self.descriptor_type:#06x} '
                f'#{self.descriptor_id}')

    def get_chain(self, *, follow_chain=True):
        """Compose a value representing this descriptor.

        With all references resolved and all chains condensed (if follow_chain
        is True), else just return THIS descriptor's info.
        """
        res = self.get_info()
        if self.FIELDS[0][1] == 'prev' and follow_chain:
            if self.prev:
                return [*self.prev.get_chain(follow_chain=follow_chain), res]
            return [res]
        return res

    def resolve(self, indexer):
        """Resolve all descriptor ID fields, except for the prev field."""
        for entry in self.FIELDS:
            if len(entry) < 4:
                continue
            _fmt, field, resolve_method, allowed_classes = entry
            value = getattr(self, field)
            if resolve_method in 'sc':
                value = self._query_chain(indexer, value, allowed_classes)
            if resolve_method == 'd':
                value = self._query_descriptor(indexer, value, allowed_classes)
            if resolve_method == 's':
                value = value.get_info() if value else ''
            setattr(self, field, value)
        self.is_resolved = True

    # helpers

    def resolve_chain(self, indexer, descr_classes):
        """Resolve our 'prev-chain' recursively.

        Checks descriptor classes on the way.
        """
        if self.FIELDS[0][1] != 'prev':
            raise RuntimeError('resolve_chain only works on descriptors '
                               'with a prev field')
        if isinstance(self.prev, Descriptor):
            return
        if not self.prev:
            self.prev = None
            return
        descr = indexer.query_descriptor(self.prev)
        if descr.descriptor_type >> 12 not in descr_classes:
            raise DescError(self, f'prev descriptor {descr} has unexpected '
                            'class, expected one of {descr_classes}')
        self.prev = descr
        # pylint: disable=protected-access
        descr.resolve_chain(indexer, descr_classes)

    def _query_chain(self, indexer, descr_id, descr_classes):
        """Query a chain of descriptors, returning the last descriptor."""
        if not descr_id:
            return None
        descr = indexer.query_descriptor(descr_id)
        if descr.descriptor_type >> 12 not in descr_classes:
            raise DescError(self, f'chain descriptor {descr} is of wrong '
                            f'type, expected one of {descr_classes}')
        # pylint: disable=protected-access
        descr.resolve_chain(indexer, descr_classes)
        return descr

    def _query_descriptor(self, indexer, descr_id, descr_classes):
        """Query a single descriptor, without building a chain."""
        if not descr_id:
            return None
        descr = indexer.query_descriptor(descr_id)
        if descr.descriptor_type >> 12 not in descr_classes:
            raise DescError(self, f'linked descriptor {descr} is of wrong '
                            f'type, expected one of {descr_classes}')
        return descr

    # methods to be overridden in derived descriptors

    def decode(self):
        """Decode packed/encoded information in the descriptor data.

        Typically rewrites attributes on self with the decoded data.
        """

    def get_info(self):
        """Return a high-level repr of this descriptor as a dictionary."""
        raise NotImplementedError


class PLCDescriptor(Descriptor):
    FIELDS = (('H', 'last_device'),
              ('H', 'description', 's', {DESCRIPTOR_CLASS_STRING}),
              ('H', 'version',     's', {DESCRIPTOR_CLASS_STRING}),
              ('H', 'author',      's', {DESCRIPTOR_CLASS_STRING}),
              ('H', 'slot_size'),
              ('H', 'num_devices'),
              ('H', 'flags'),
              ('s', 'name'),
              )
    last_device = description = version = author = slot_size = num_devices = \
        flags = name = None

    def decode(self):
        self.num_slots = 1 + (self.flags >> 12)
        self.flags = []

    def check(self):
        if self.slot_size < 32:
            raise DescError(self, f'slot size ({self.slot_size}) too small, '
                            'should at least be 32')
        if self.slot_size % 8:
            raise DescError(self, f'slot size ({self.slot_size}) must be '
                            'divisble by 8')

    def get_info(self):
        self.check()
        return dict(
            type='plc',
            description=self.description,
            version=self.version,
            author=self.author,
            slot_size=self.slot_size,
            num_devices=self.num_devices,
            flags=self.flags,
            num_slots=self.num_slots,
            name=self.name,
        )


class DeviceDescriptor(Descriptor):
    FIELDS = (('H', 'prev'),
              ('H', 'description', 's', {DESCRIPTOR_CLASS_STRING}),
              ('H', 'main_param',  'd', {DESCRIPTOR_CLASS_PAR}),
              ('H', 'aux',         'c', {DESCRIPTOR_CLASS_BF}),
              ('H', 'accessibles', 'c', {DESCRIPTOR_CLASS_PAR,
                                         DESCRIPTOR_CLASS_FUNC,
                                         DESCRIPTOR_CLASS_TABLE}),
              ('H', 'errid',       'c', {DESCRIPTOR_CLASS_BF}),
              ('H', 'typecode'),
              ('H', 'address'),
              ('H', 'flags'),
              ('s', 'name'),
              )
    prev = description = main_param = aux = accessibles = errid = typecode = \
        address = flags = name = None

    def decode(self):
        flags = []
        for flag, text in DEVICE_FLAGS.items():
            if self.flags & (1 << flag):
                flags.append(text)
        self.flags = flags

    def check(self):
        if self.address % 2:
            raise DescError(self, f'address {self.address} should be at '
                            'least even')
        # XXX: check chain of accessibles

    def get_info(self):
        self.check()
        # fixup main_param
        main_param = self.main_param.get_info() if self.main_param else {}
        main_param.pop('type', None)
        main_param.pop('idx', None)
        main_param.pop('name', None)
        main_param.pop('description', None)

        mp_basetype = main_param.pop('basetype', None)
        mp_width = main_param.pop('width', 0)
        # pylint: disable=import-outside-toplevel
        from zapf.device import TYPECODE_MAP  # noqa: PLC0415
        try:
            typecode_info = TYPECODE_MAP[self.typecode]
            basetype, width = FMT_TO_BASETYPE[typecode_info.value_fmt]
            readonly = typecode_info.readonly
            # correct basetype if param descriptor is enum
            if mp_basetype == 'enum' and basetype in ('int', 'uint', 'float'):
                basetype = 'enum'
            if mp_width and mp_width != width:
                raise DescError(self, f'width {mp_width} declared in main '
                                'param descriptor does not match typecode')
        except KeyError:
            basetype, width, readonly = None, 0, True
        access = 'ro' if readonly or 'readonly' in self.flags else 'rw'
        absmin = main_param.pop('min_value', -FLOAT32_MAX)
        absmax = main_param.pop('max_value', FLOAT32_MAX)
        unit = main_param.pop('unit', '')
        enum_r = main_param.pop('enum_r', None)
        enum_w = main_param.pop('enum_w', None)
        params, funcs, tables = {}, {}, {}
        descr = self.accessibles

        # XXX: need stricter checking of alloced ressources (IDX, mem area, etc...)
        def fixup(entry):
            # fix unit
            if 'unit' in entry:
                entry['unit'] = entry['unit'].replace('#', unit)
            return entry

        while descr:
            if isinstance(descr, (NumericParameterDescriptor, EnumParameterDescriptor)):
                if descr.name not in params:
                    entry = descr.get_info()
                    name = entry.pop('name')
                    params[name] = fixup(entry)
            elif isinstance(descr, SpecialFunctionDescriptor):
                if descr.name not in funcs:
                    entry = descr.get_info()
                    name = entry.pop('name')
                    # fix unit
                    d = entry['argument']
                    if d:
                        entry['argument'] = fixup(d)
                    d = entry['result']
                    if d:
                        entry['argument'] = fixup(d)
                    funcs[name] = entry
            elif isinstance(descr, TableDescriptor):
                if descr.name not in tables:
                    entry = descr.get_info()
                    name = entry.pop('name')
                    tables[name] = fixup(entry)
            else:
                raise DescError(self, 'accessible chain must only contain param/'
                                f'function/table descriptors, not {descr}')
            descr = descr.prev

        # clean bitfield descriptions for aux/errid, if existing
        def clean_bitfield_chain(chain):
            res = []
            seen = {}
            for elem in reversed(chain):
                if len(elem) == 2:  # flag: bit, name
                    if elem[0] not in seen:
                        seen[elem[0]] = elem
                        res.append(elem)
                elif len(elem) == 4:  # bitfield: lsb, width, name, enum
                    # if any bit of the bitfield already seen -> skip bitfield
                    for i in range(elem[0], elem[0] + elem[1]):
                        if i not in seen:
                            continue
                        newer = seen[i]
                        if len(newer) == 4:  # also a bitfield
                            # -> update newer with the older entries, IF THEY FIT
                            # same startbit and newer is at least as wide
                            if newer[0] == elem[0] and newer[1] >= elem[1]:
                                for k, v in elem[3].items():  # iter over the enums
                                    # update if not already given later
                                    newer[3].setdefault(k, v)
                            res.remove(newer)
                            res.append(newer)
                        break
                    else:
                        for i in range(elem[0], elem[0] + elem[1]):
                            seen[i] = elem
                        res.append(elem)
            # now sort res ascending
            res.sort()
            return res

        aux = clean_bitfield_chain(self.aux.get_chain()) if self.aux else []
        errid = clean_bitfield_chain(self.errid.get_chain()) if self.errid else []
        # TODO: need a way to check/complain if a parameterless device has
        # parameter descriptors, also main parameter type of MessageIO
        if access == 'ro' and enum_w:
            enum_w = {}
        if basetype == 'enum' and (enum_r or enum_w):
            enum_values = set(enum_r.keys()).union(set(enum_w.values()))
            absmin = min(enum_values)
            absmax = max(enum_values)
        if basetype in ('int', 'uint'):
            absmin = int(absmin)
            absmax = int(absmax)

        return dict(
            description=self.description,
            aux=aux,
            params=params,
            funcs=funcs,
            tables=tables,
            errid=errid,
            typecode=self.typecode,
            address=self.address,
            flags=self.flags,
            absmin=absmin,
            absmax=absmax,
            unit=unit,
            basetype=basetype,
            width=width,
            access=access,
            enum_r=enum_r,
            enum_w=enum_w,
            name=self.name,
        )


class StringDescriptor(Descriptor):
    FIELDS = (('H', 'prev'),
              ('s', 'text'),
              )
    prev = text = None

    def get_info(self):
        if self.prev:
            if not isinstance(self.prev, StringDescriptor):
                raise DescError(self, 'string descriptors may not reference '
                                'other types of descriptors ({self.prev})')
            return self.prev.get_info() + self.text
        return self.text


class EnumDescriptor(Descriptor):
    FIELDS = (('H', 'prev'),
              ('H', 'val'),
              ('s', 'name'),
              )
    prev = val = name = None

    def get_info(self):
        if self.prev:
            if not isinstance(self.prev, EnumDescriptor):
                raise DescError(self, 'enum descriptors may not reference '
                                'other types of descriptors ({self.prev})')
            res = self.prev.get_info()
            res.update({self.val: self.name})
            return res
        return {self.val: self.name}


class BitfieldDescriptor(Descriptor):
    FIELDS = (('H', 'prev'),
              ('H', 'enum', 'c', {DESCRIPTOR_CLASS_ENUM}),
              ('B', 'lsb'),
              ('B', 'width'),
              ('s', 'name'),
              )
    prev = enum = lsb = width = name = None

    def decode(self):
        self.name = self.name.replace(' ', '_')

    def check(self):
        if self.width < 1:
            raise DescError(self, 'width of bitfield must be at least 1')
        mask = (1 << self.width) - 1
        if self.enum and max(self.enum.get_info()) > mask:
            raise DescError(self, "enum chain contains entries which won't "
                            f'fit into {self.width} bits')
        if self.prev and not isinstance(self.prev, (BitfieldDescriptor,
                                                    FlagDescriptor)):
            raise DescError(self, 'bitfield descriptors may only reference '
                            f'bitfield or flag descriptors, not {self.prev}')
        # XXX: check overrides correctly!

    def get_info(self):
        self.check()
        # XXX: handle overrides correctly!
        return (self.lsb, self.width, self.name,
                self.enum.get_info() if self.enum else {})


class FlagDescriptor(Descriptor):
    FIELDS = (('H', 'prev'),
              ('B', 'bit'),
              ('s', 'name'),
              )
    prev = bit = name = None

    def get_info(self):
        if self.prev and not isinstance(self.prev, (BitfieldDescriptor,
                                                    FlagDescriptor)):
            raise DescError(self, 'flag descriptors may only reference '
                            f'bitfield or flag descriptors, not {self.prev}')
        return (self.bit, self.name)


class NumericParameterDescriptor(Descriptor):
    FIELDS = (('H', 'prev'),
              ('H', 'description', 'c', {DESCRIPTOR_CLASS_STRING}),
              ('H', 'idx'),
              ('H', 'ptype'),
              ('H', 'unit'),
              ('f', 'min_value'),
              ('f', 'max_value'),
              ('s', 'name'),
              )
    prev = description = idx = ptype = unit = min_value = max_value = name = None

    def decode(self):
        self.unit = decode_unit(self.unit)
        self.idx += (self.ptype & 0xFF) << 16
        self.basetype = {0: 'uint', 1: 'float', 3: 'int'}[(self.ptype >> 14) & 0x3]
        self.access = {1: 'rw', 2: 'ro', 3: 'obs'}[(self.ptype >> 12) & 0x3]
        self.width = {0: 0, 1: 16, 2: 32, 3: 64}[(self.ptype >> 10) & 0x3]

    def check(self):
        if self.max_value < self.min_value:
            raise DescError(self, f'parameter max_value ({self.max_value}) must '
                            f'be bigger than min_value ({self.min_value})')
        if self.prev and \
           isinstance(self.prev, Descriptor) and \
           not isinstance(self.prev, (NumericParameterDescriptor,
                                      EnumParameterDescriptor,
                                      SpecialFunctionDescriptor,
                                      TableDescriptor)):
            raise DescError(self, 'param descriptors may only reference '
                            f'param/function/table descriptors, not {self.prev}')

    def get_info(self):
        self.check()
        return dict(
            type='param',
            description=self.description.get_info() if self.description else '',
            idx=self.idx,
            basetype=self.basetype,
            width=self.width,
            access=self.access,
            unit=self.unit,
            min_value=self.min_value,
            max_value=self.max_value,
            name=self.name,
        )


class EnumParameterDescriptor(Descriptor):
    FIELDS = (('H', 'prev'),
              ('H', 'description', 'c', {DESCRIPTOR_CLASS_STRING}),
              ('H', 'enum_r',      'c', {DESCRIPTOR_CLASS_ENUM}),
              ('H', 'enum_w',      'c', {DESCRIPTOR_CLASS_ENUM}),
              ('H', 'idx'),
              ('H', 'ptype'),
              ('s', 'name'),
              )
    prev = description = enum_r = enum_w = idx = ptype = name = None

    def decode(self):
        self.idx += (self.ptype & 0xFF) << 16
        self.width = {0: 0, 1: 16, 2: 32, 3: 64}[(self.ptype >> 10) & 0x3]
        self.access = {1: 'rw', 2: 'ro', 3: 'obs'}[(self.ptype >> 12) & 0x3]
        if not self.enum_r:
            self.enum_r = self.enum_w
        if not self.enum_w:
            self.enum_w = self.enum_r

        if not self.enum_r:
            raise DescError(self, 'enums for an enum parameter may not '
                            'be empty')

    def check(self):
        enum_r = self.enum_r.get_info() if self.enum_r else {}
        enum_w = self.enum_w.get_info() if self.enum_w else {}
        if (self.ptype >> 14) & 0x3 != 2:
            raise DescError(self, 'enum descriptor parameter ptype needs '
                            f'to be enum, not {self.ptype}')
        for k, v in enum_w.items():
            if k not in enum_r:
                raise DescError(self, 'all write_enum entries must also be '
                                f'in the read_enums (missing {k!r})')
            if enum_r[k] != v:
                raise DescError(self, 'all write_enum entries must be the '
                                'same as in the read_enums (error at {k!r})')
        if self.prev and not isinstance(self.prev, (NumericParameterDescriptor,
                                                    EnumParameterDescriptor,
                                                    SpecialFunctionDescriptor,
                                                    TableDescriptor)):
            raise DescError(self, 'param descriptors may only reference '
                            f'param/function/table descriptors, not {self.prev}')

    def get_info(self):
        self.check()
        # reverse enum_w
        enum_w = {}
        for k, v in (self.enum_w.get_info() if self.enum_w else {}).items():
            enum_w[v] = k  # noqa: PERF403
        return dict(
            type='param',
            description=self.description.get_info() if self.description else '',
            idx=self.idx,
            basetype='enum',
            width=self.width,
            access=self.access,
            enum_r=self.enum_r.get_info() if self.enum_r else {},
            enum_w=enum_w,
            name=self.name,
        )


class SpecialFunctionDescriptor(Descriptor):
    FIELDS = (('H', 'prev'),
              ('H', 'description', 'c', {DESCRIPTOR_CLASS_STRING}),
              ('H', 'argument',    'd', {DESCRIPTOR_CLASS_PAR}),
              ('H', 'result',      'd', {DESCRIPTOR_CLASS_PAR}),
              ('H', 'idx'),
              ('H', 'ptype'),
              ('s', 'name'),
              )
    prev = description = argument = result = idx = ptype = name = None

    def decode(self):
        self.idx += (self.ptype & 0xFF) << 16

    def check(self):
        # XXX: what to check?
        pass

    def get_info(self):
        self.check()
        if (self.ptype >> 12) & 0x3 != 0:
            raise DescError(self, 'function descriptors must have ptype = 0, '
                            f'not {self.ptype}')
        if self.prev and not isinstance(self.prev, (NumericParameterDescriptor,
                                                    EnumParameterDescriptor,
                                                    SpecialFunctionDescriptor,
                                                    TableDescriptor)):
            raise DescError(self, 'function descriptors may only reference '
                            f'param/function/table descriptors, not {self.prev}')
        if self.argument and not isinstance(self.argument, (NumericParameterDescriptor,
                                                            EnumParameterDescriptor)):
            raise DescError(self, 'argument descriptors must be param '
                            f'descriptors, not {self.argument}')
        if self.result and not isinstance(self.result, (NumericParameterDescriptor,
                                                        EnumParameterDescriptor)):
            raise DescError(self, 'result descriptors must be one param '
                            f'descriptors, not {self.result}')
        arg = self.argument.get_info() if self.argument else {}
        arg.pop('type', None)
        arg.pop('idx', None)
        res = self.result.get_info() if self.result else {}
        res.pop('type', None)
        res.pop('idx', None)
        return dict(
            description=self.description.get_info(),
            argument=arg,
            result=res,
            idx=self.idx,
            name=self.name,
        )


class TableDescriptor(Descriptor):
    FIELDS = (('H', 'prev'),
              ('H', 'description', 'c', {DESCRIPTOR_CLASS_STRING}),
              ('H', 'header',      'c', {DESCRIPTOR_CLASS_PAR, DESCRIPTOR_CLASS_FUNC}),
              ('H', 'idx'),
              ('H', 'flags'),
              ('H', 'last_row'),
              ('H', 'last_column'),
              ('s', 'name'),
              )
    prev = description = header = idx = flags = last_row = last_column = name = None

    def decode(self):
        self.accesstype = {
            0: TABLE_ACCESS_PCTL,
            1: TABLE_ACCESS_MEM_1D,
            2: TABLE_ACCESS_MEM_2D,
        }[self.flags >> 14]

    def check(self):
        # XXX: check chain of header descr.
        if self.prev and not isinstance(self.prev, (NumericParameterDescriptor,
                                                    EnumParameterDescriptor,
                                                    SpecialFunctionDescriptor,
                                                    TableDescriptor)):
            raise DescError(self, 'table descriptors may only reference '
                            f'param/function/table descriptors, not {self.prev}')

    def get_info(self):
        self.check()
        header = self.header.get_chain()  # keep as list to keep the order!
        # overriding logic (headerlist is already reversed)
        # + check header entries
        seen = {}
        undefined_columns = set(range(self.last_column + 1))
        for entry in header:
            # check bad idx
            if entry['idx'] > self.last_column:
                # or just ignore those???
                raise SpecError(f'table {self.name!r} header chain entry '
                                f'{entry["name"]!r} has idx '
                                f'{entry["idx"]} > {self.last_column}!')
            # overrides 'in-order' by 'idx' field (as spec say)
            seen[entry['idx']] = entry
            undefined_columns.discard(entry['idx'])
        if undefined_columns:
            raise SpecError(f'table {self.name!r} header chain contains '
                            f'no entry/ies for column(s) '
                            f'{", ".join(str(d) for d in undefined_columns)}')
        header = list(seen.values())
        return dict(
            type='table',
            description=self.description.get_info(),
            header=header,
            idx=self.idx,
            flags=self.flags,
            accesstype=self.accesstype,
            last_row=self.last_row,
            last_column=self.last_column,
            name=self.name,
        )


class DebugDescriptor(Descriptor):
    FIELDS = (('H', 'cycle'),
              ('H', 'indexer_size'),
              ('s', 'text'),
              )
    cycle = text = None

    def get_info(self):
        return f'{self.cycle:#06x}: {self.text!r}'


# Note: the fmt for the trailing string is computed and appended
# in indexer.query_descriptor
DESCRIPTOR_TYPES = {  # map DescriptorType to minstorage
    0x10: PLCDescriptor,
    0x20: DeviceDescriptor,
    0x30: StringDescriptor,
    0x40: EnumDescriptor,
    0x50: BitfieldDescriptor,
    0x51: FlagDescriptor,
    0x61: NumericParameterDescriptor,
    0x62: EnumParameterDescriptor,
    0x68: SpecialFunctionDescriptor,
    0x6C: TableDescriptor,
    0xFF: DebugDescriptor,
}

class ValueInfo(NamedTuple):
    """Info about a value (device value, parameter, sfunc arg/result, table column)."""

    basetype: str
    fmt: str
    readonly: bool
    unit: str
    min_value: float
    max_value: float
    enum_r: dict | None
    enum_w: dict | None
    description: str


def get_valueinfo(entry, device_valuesize):
    key = (entry['basetype'], entry['width'] or 8*device_valuesize)
    fmt, min_value, max_value = BASETYPE_TO_FMT[key]
    if struct.calcsize(fmt) > device_valuesize:
        raise SpecError('no parameter can have a bigger width than the main '
                        'value of a device')
    return ValueInfo(entry['basetype'],
                     fmt,
                     entry['access'] != 'rw',
                     entry.get('unit', ''),
                     entry.get('min_value', min_value),
                     entry.get('max_value', max_value),
                     entry.get('enum_r', None),
                     entry.get('enum_w', None),
                     entry.get('description', ''),
                     )


BASE_UNITS = {
    0: '',
    1: '#',
    2: 'bar',
    3: 'counts',
    4: 'deg',
    5: 'g',
    6: 'm',
    7: 'm^2',
    8: 'm^3',
    9: 's',
    10: 'A',
    11: 'F',
    12: 'H',
    13: 'K',
    14: 'Ohm',
    15: 'T',
    16: 'V',
    17: 'W',
    18: 'degC',
    19: 'degF',
    20: 'bit',
    21: 'steps',
    22: 'hours',
}

TIME_BASE = {
    0: '',
    1: '/(#)',
    2: '/s',
    3: '/s^2',
    4: '/s^3',
    5: '/min',
    6: '/h',
    7: '/d',
}

EXPONENTS = {
    0: '',
    2: 'h',
    3: 'k',
    6: 'M',
    9: 'G',
    12: 'T',
    15: 'P',
    -1: 'd',
    -2: 'c',
    -3: 'm',
    -6: 'u',
    -9: 'n',
    -12: 'f',
    -15: 'a',
}

SPECIALS = [  # final decoded, from lookup, exponent correction
    ('l', 'm^3', -3),  # 'l' is 10^-3 m^3
    ('%', '', -2),     # '%' is 10^-2
]


# for tests/testplc
def encode_unit(unit=''):
    if unit == '':
        return 0
    unit = unit.strip()
    for s in SPECIALS:
        if s[0] in unit:
            return (encode_unit(unit.replace(s[0], s[1])) + (s[2] << 11)) & 0xFFFF
    timebase = 0
    if '/' in unit:
        for tbc, tbv in TIME_BASE.items():
            if tbv and unit.endswith(tbv):
                timebase = tbc
                unit = unit.rsplit(tbv, 1)[0]
                unit = unit.strip()
                break
        else:
            raise ApiError(f'can\'t encode unit "{unit}"')
    baseunit = 0
    if unit.endswith('1'):
        unit = unit[1:]
    if unit:
        for bc, bv in BASE_UNITS.items():
            if bv and unit.endswith(bv):
                baseunit = bc
                unitprefix = unit.rsplit(bv, 1)[0]
                unitprefix = unitprefix.strip()
                for ec, ev, in EXPONENTS.items():
                    if unitprefix == ev:
                        return (baseunit + (timebase << 6) + (ec << 11)) & 0xFFFF
                for ec in range(-16, 16):
                    if unitprefix == f'10^{ec}':
                        return (baseunit + (timebase << 6) + (ec << 11)) & 0xFFFF
        raise ApiError(f'can\'t encode unit "{unit}"')
    for ec, ev, in EXPONENTS.items():
        if unit == ev:
            return (baseunit + (timebase << 6) + (ec << 11)) & 0xFFFF
    for ec in range(-16, 16):
        if unit == f'10^{ec}':
            return (baseunit + (timebase << 6) + (ec << 11)) & 0xFFFF
    raise ApiError(f'can\'t encode unit "{unit}"')


def decode_unit(code):
    base_code = code & 0x3F
    tb_code = (code >> 6) & 0x1F
    exp_code = (code >> 11) & 0x1F
    if exp_code > 15:
        # make signed
        exp_code -= 32

    if base_code == tb_code == 1:
        base_code = tb_code = 0

    base_unit = BASE_UNITS.get(base_code, 'unknown unit')
    for s in SPECIALS:
        if base_unit == s[1] and exp_code <= s[2]:
            base_unit = s[0]
            exp_code -= s[2]
    time_base = TIME_BASE.get(tb_code, '')
    if time_base and not base_unit:
        return (f'10^{exp_code}' if exp_code else '1') + time_base
    exponent = f'10^{exp_code} ' if exp_code else ''
    if '^' not in base_unit and base_code > 1:
        exponent = EXPONENTS.get(exp_code, exponent)
    return f'{exponent}{base_unit}{time_base}'


def fix_unit(param_unit, main_unit):
    # fixes the unit strings
    if '#' not in param_unit:
        return param_unit
    if not main_unit:
        if '1/(#)' in param_unit:
            return param_unit.replace('1/(#)', '')
        if '#' in param_unit:
            return param_unit.replace('#', '')
    return param_unit.replace('#', main_unit)



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Contains all necessary constants and mappings from the spec for 2015.02 only."""

# 2015.02
from zapf.spec import FLOAT32_MAX

INDEXER_DEV = 0

INFO_STRUCT      = 0
INFO_SIZE        = 1
INFO_ADDR        = 2
INFO_UNIT        = 3
INFO_NAME        = 4
INFO_VERSION     = 5
INFO_AUTHOR1     = 6
INFO_AUTHOR2     = 7
INFO_DESCRIPTOR  = 8
INFO_PARAMS      = 15
INFO_AUX1        = 16
INFO_CYCLE       = 127


UNIT_CODES    = ('', 'V', 'A', 'W', 'm', 'g', 'Hz', 'T', 'K', 'degC', 'degF',
                 'bar', 'deg', 'Ohm', 'm/s', 'm^2/s', 'm^3/s', 's', 'cts',
                 'bar/s', 'bar/s^2', 'F', 'H', 'l/min')
UNIT_EXPONENT = {0: '', 2: 'h', 3: 'k', 6: 'M', 9: 'G', 12: 'T', 15: 'P',
                 18: 'E', -1: 'd', -2: 'c', -3: 'm', -6: 'u', -9: 'n',
                 -12: 'f', -15: 'a'}
UNIT_SPECIAL  = {(0, -2): '%', (16, -3): 'l/s'}


def convert_unit(code, exponent):
    if (code, exponent) in UNIT_SPECIAL:
        return UNIT_SPECIAL[code, exponent]
    try:
        unit = UNIT_CODES[code]
    except IndexError:
        unit = 'unit'
    if '^' in unit:
        return f'10^{exponent} {unit}'
    return UNIT_EXPONENT.get(exponent, f'10^{exponent} ') + unit


# The first (numbered) parameter whose value is float instead of integer.
FIRST_FLOAT_PARAM = 30


def is_function(idx):
    return 128 <= idx < 192 or 224 <= idx < 240


# helpers for converting a 2015.02 into a 2021.09 spec
PARAMETER_Specs = {
    1:  dict(name='Mode', basetype='int'),
    2:  dict(name='MicroSteps', basetype='int', minval=1),
    3:  dict(name='CoderBits', basetype='int', minval=1),
    4:  dict(name='LegacyExtStatus', basetype='int'),
    5:  dict(name='LegacyHwOffset', basetype='int'),

    30: dict(name='LegacyAbsMin', unit='#'),
    31: dict(name='LegacyAbsMax', unit='#'),
    32: dict(name='UserMin', unit='#'),
    33: dict(name='UserMax', unit='#'),
    34: dict(name='WarnMin', unit='#'),
    35: dict(name='WarnMax', unit='#'),
    36: dict(name='TimeoutTime', minval=0, unit='s'),
    37: dict(name='MaxTravelDist', minval=0, unit='#'),
    38: dict(name='AccelTime', minval=0, unit='s'),
    40: dict(name='Offset', unit='#'),
    43: dict(name='BlockSize', minval=0, unit='#'),
    44: dict(name='Opening', minval=0, unit='#'),
    51: dict(name='PidP', minval=0, maxval=100, unit='%/(#)'),
    52: dict(name='PidI', minval=0, maxval=900, unit='s'),
    53: dict(name='PidD', minval=0, maxval=100, unit='1/s'),
    55: dict(name='DragError', minval=0, unit='#'),
    56: dict(name='Hysteresis', minval=0, unit='#'),
    57: dict(name='Holdback', minval=0, unit='#'),
    58: dict(name='HomingSpeed', minval=0, unit='#/s'),
    59: dict(name='Jerk', minval=0, unit='#/s'),
    60: dict(name='Speed', minval=0, unit='#/s'),
    61: dict(name='Accel', minval=0, unit='#/s^2'),
    62: dict(name='IdleCurrent', minval=0, maxval=100, unit='A'),
    63: dict(name='RampCurrent', minval=0, maxval=100, unit='A'),
    64: dict(name='MoveCurrent', minval=0, maxval=100, unit='A'),
    65: dict(name='StopCurrent', minval=0, maxval=100, unit='A'),
    66: dict(name='LegacyAbortDecel'),
    67: dict(name='LegacyMicrosteps'),
    68: dict(name='Slope', unit='steps/(#)'),
    69: dict(name='HomePosition', unit='#'),
    70: dict(name='Setpoint', access='obs', unit='#'),
}

# special functions
FUNC_Specs = {
    128: dict(name='FactoryReset', argument={'basetype': 'int'}),
    130: dict(name='LegacyAbort'),
    131: dict(name='LegacyHomeNeg'),
    132: dict(name='LegacyHomePos'),
    133: dict(name='Home'),
    137: dict(name='SetPosition', argument={'unit': '#'}),
    142: dict(name='ContMove', argument={'unit': '#/s'}),
}

# fill up undef'd + custom idx with generic entries (param224, etc..)
for i in range(1, 240):
    if not is_function(i):
        if i not in PARAMETER_Specs:
            PARAMETER_Specs[i] = dict(name=f'Param{i}')
    elif i not in FUNC_Specs:
        FUNC_Specs[i] = dict(name=f'Command{i}')
for i in range(240, 256):
    PARAMETER_Specs[i] = dict(name=f'Reserved{i}')

# augment with default values
for v in PARAMETER_Specs.values():
    v.setdefault('basetype', 'float')
    def_minval, def_maxval = (-FLOAT32_MAX, FLOAT32_MAX) \
        if v['basetype'] == 'float' else (-(2 ** 31), 2 ** 31 - 1)
    v.setdefault('unit', '')
    v.setdefault('min_value', def_minval)
    v.setdefault('max_value', def_maxval)
    v['width'] = 0  # ALWAYS same as device value
    v.setdefault('access', 'rw')
    v.setdefault('type', 'param')
    v.setdefault('description', '')

for v in FUNC_Specs.values():
    if 'argument' in v:
        arg = v['argument']
        arg.setdefault('basetype', 'float')
        arg['width'] = 0
        arg['access'] = 'rw'
        arg.setdefault('unit', '')
        arg.setdefault('min_value', -FLOAT32_MAX)
        arg.setdefault('max_value', FLOAT32_MAX)
        v.setdefault('type', 'param')
    v.setdefault('argument', None)
    v.setdefault('result', None)  # does not occur
    v.setdefault('description', '')



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Interaction with the indexer."""

import time

from zapf import CommError, SpecError, spec
from zapf.spec import v_2015_02, v_2021_09


class Indexer:
    IO_RETRIES = 32

    def __init__(self, io, log):
        self.io = io
        self.log = log
        self.magicstr = None
        self.indexer_addr = None
        self.indexer_size = None
        self.num_devices = 0
        self.plc_name = None
        self.plc_version = None
        self.plc_author = None
        self.plc_description = None
        self._descr_cache = {}
        self.error = None

    def detect_plc(self, extended=True):  # noqa: FBT002
        """Check if the HW follows the spec and which MAGIC number it has.

        Afterwards, checks the indexer and reads extended meta information
        about the PLC if wanted.
        """
        if self.indexer_addr and self.indexer_size and self.plc_author:
            return

        # check MAGIC
        magic = self.io.detect_magic()
        magicstr = f'{magic:.2f}'.replace('.', '_')
        if magicstr not in spec.SUPPORTED_MAGICS:
            raise SpecError(f'magic {magicstr} is not supported by this client')
        self.magicstr = magicstr

        # read indexer offset from memory address 4
        addr = self.io.read_u16(spec.OFFSET_ADDR)
        if addr < 6 or addr & 1:
            raise SpecError(f'indexer offset {addr} is invalid')
        self.indexer_addr = addr

        # continue in a magic specific way
        getattr(self, f'_check_{magicstr}')(extended)

    def _check_2014_07(self, extended):
        self.log.debug('checking a 2014.07 PLC from the stone age....')
        self.indexer_slot_size = 2
        self.indexer_num_slots = 1
        self.indexer_data_size = self.indexer_slot_size * self.indexer_num_slots
        self.indexer_size = 2 + self.indexer_data_size

        self.num_devices = 0

        # query firmware information
        if extended:
            self.plc_name = 'crap'
            self.plc_version = 'crap'
            self.plc_author = 'Anonymous Coward'
            self.plc_description = 'crappy 2014.07 PLC'

    def _check_2015_02(self, extended):
        # query indexer size
        size = self.query_word(v_2015_02.INDEXER_DEV, v_2015_02.INFO_SIZE)
        if size < 22 or size > 66 or size & 1:
            raise SpecError(f'indexer size {size} is invalid')
        self.indexer_slot_size = size - 2
        self.indexer_num_slots = 1
        self.indexer_data_size = self.indexer_slot_size * self.indexer_num_slots
        self.indexer_size = 2 + self.indexer_data_size

        # query indexer size and offset again, through the info struct,
        # and ensure consistency
        info = self.query_infostruct(v_2015_02.INDEXER_DEV)
        self.num_devices = 0
        if info[0] + info[1] + info[2] != 0:
            if info[0] != 0 or \
               info[1] != self.indexer_size or \
               info[2] != self.indexer_addr:
                raise SpecError('indexer information from infostruct does not '
                                'match with OFFSET or size')

            # total number of devices can be given in flags
            self.num_devices = info[4] & 0xFF

        # query firmware information
        if extended:
            self.plc_name = self.query_string(
                v_2015_02.INDEXER_DEV, v_2015_02.INFO_NAME,
            )
            self.plc_version = self.query_string(
                v_2015_02.INDEXER_DEV, v_2015_02.INFO_VERSION,
            )
            author1 = self.query_string(v_2015_02.INDEXER_DEV, v_2015_02.INFO_AUTHOR1)
            author2 = self.query_string(v_2015_02.INDEXER_DEV, v_2015_02.INFO_AUTHOR2)
            self.plc_author = (author1 or 'Anonymous') + '\n' + author2
            self.plc_description = ''

    def _check_2021_09(self, extended):
        # safe starting values (mailbox + minimum slot size)
        self.indexer_slot_size = 32
        self.indexer_num_slots = 1
        self.indexer_data_size = 32
        self.indexer_size = 2 + self.indexer_data_size

        # read beginning of PLC-descripor
        initial_plc_descriptor = self.query_descriptor_uncached(0)

        error = False
        # in case the indexer is stuck at the DebugDescriptor:
        if isinstance(initial_plc_descriptor, v_2021_09.DebugDescriptor):
            self.indexer_slot_size = initial_plc_descriptor.indexer_size
            error = True
        else:
            self.indexer_slot_size = initial_plc_descriptor.slot_size
            self.indexer_num_slots = initial_plc_descriptor.num_slots

        # set correct values
        self.indexer_data_size = self.indexer_slot_size * self.indexer_num_slots
        self.indexer_size = 2 + self.indexer_data_size

        # re-read full sized plc_descriptor
        plc_descriptor = self.query_descriptor(v_2021_09.DEBUG_Descriptor
                                               if error else 0)
        if error:
            self.error = self.plc_description = plc_descriptor.text
            raise CommError('can not initialize, PLC stuck at DebugDescriptor '
                            f'{plc_descriptor.text!r}')

        self.num_devices = plc_descriptor.num_devices

        # query firmware information
        if extended:
            info = plc_descriptor.get_info()
            self.plc_name = info['name']
            self.plc_version = info['version']
            self.plc_author = info['author'] or 'Anonymous Author'
            self.plc_description = info['description']

        self.plc_descriptor = plc_descriptor

    # lowlevel methods to query the indexer

    def query_infostruct(self, devnum):
        # fields in the infostruct: typecode, size, addr, unitcode, unitexp,
        # flags, absmin, absmax, the rest is the name
        result = self.query_data(
            devnum, v_2015_02.INFO_STRUCT, f'HHHBbIII{self.indexer_size - 22}s',
        )
        # convert min/max; these are floats, but we need to potentially word-
        # swap them
        absmin = self.io.float_from_dword(result[6])
        absmax = self.io.float_from_dword(result[7])
        # convert unit to a string
        unit = v_2015_02.convert_unit(*result[3:5])
        # only use the name if it has a trailing null byte, so we can be
        # sure it was fully transferred in the reduced space at the end
        name_parts = result[-1].partition(b'\0')
        name = name_parts[0].decode('latin1') if name_parts[1] else ''
        return (*result[:3], unit, result[5], absmin, absmax, name)

    def query_word(self, devnum, infotype):
        return self.query_data(devnum, infotype, 'H')[0]

    def query_unit(self, devnum, infotype):
        return v_2015_02.convert_unit(*self.query_data(devnum, infotype, 'Bb'))

    def query_string(self, devnum, infotype):
        result = self.query_bytes(devnum, infotype)
        return result.partition(b'\0')[0].decode('latin1')

    def query_bitmap(self, devnum, infotype):
        return [
            grp * 8 + bit
            for grp, byte in enumerate(self.query_bytes(devnum, infotype))
            for bit in range(8)
            if byte & (1 << bit)
        ]

    def query_bytes(self, devnum, infotype):
        return self.query_data(devnum, infotype, f'{self.indexer_size - 2}s')[0]

    # new for 2021_09

    def _descriptor_from_data(self, descriptorid, descriptor_type, data):
        try:
            descr_class = v_2021_09.DESCRIPTOR_TYPES[descriptor_type >> 8]
        except KeyError:
            raise SpecError('got unsupported descriptor_type '
                            f'{descriptor_type:#06x} for descriptor id '
                            f'{descriptorid:#06x}') from None
        return descr_class(descriptorid, descriptor_type, data)

    def query_descriptor(self, desc_id):
        if desc_id in self._descr_cache:
            wanted = self._descr_cache[desc_id]
            if not wanted.is_resolved:
                wanted.resolve(self)
            return wanted

        if desc_id == v_2021_09.DEBUG_Descriptor:
            return self.query_descriptor_uncached(v_2021_09.DEBUG_Descriptor)

        # query descriptors in full blocks of #slots
        datalen = self.indexer_slot_size - 2
        first_desc_id = (desc_id // self.indexer_num_slots) * \
            self.indexer_num_slots
        slot_data = self.query_data(first_desc_id, 0,
                                    self.indexer_num_slots * f'H{datalen}s')
        cur_desc_id = first_desc_id - 1
        while slot_data:
            desc_type, data = slot_data[:2]
            slot_data = slot_data[2:]
            cur_desc_id += 1

            if desc_type == 0:
                continue

            descr = self._descriptor_from_data(cur_desc_id, desc_type, data)
            self._descr_cache[cur_desc_id] = descr

        wanted = self._descr_cache[desc_id]
        wanted.resolve(self)
        return wanted

    def query_descriptor_uncached(self, desc_id):
        if desc_id == v_2021_09.DEBUG_Descriptor:
            datalen = self.indexer_slot_size * self.indexer_num_slots - 2
        else:
            datalen = self.indexer_slot_size - 2
        desc_type, data = self.query_data(desc_id, 0, f'H{datalen}s')
        if desc_type == 0:
            return None
        return self._descriptor_from_data(desc_id, desc_type, data + b'\0')

    # even lower level methods to do an indexer transaction

    def query_data(self, devnum, infotype, fmt):
        request = infotype << 8 | devnum
        self.io.write_u16s(self.indexer_addr, [request])
        for i in range(self.IO_RETRIES):
            reply = self.io.read_fmt(self.indexer_addr, 'H' + fmt)
            if reply[0] == request | 0x8000:
                return reply[1:]
            # indexer stuck at debug descriptor?
            if reply[0] == 0xFFFF and self.magicstr == '2021_09':
                return reply[1:]
            # refresh the query if it got overwritten by a different host
            if reply[0] & 0x7FFF != request:
                self.io.write_u16s(self.indexer_addr, [request])
            time.sleep(0.001 * i * i)
        raise CommError('indexer not responding in time')



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Basic device abstraction for PILS devices."""

import time
from struct import calcsize
from typing import NamedTuple

from zapf import ApiError, CommError, SpecError
from zapf.spec import (
    DevStatus,
    ParamCMDs,
    ParamControl,
    StatusStruct16,
    StatusStruct32,
    decode_bitfield,
)
from zapf.spec.v_2021_09 import get_valueinfo
from zapf.table import Table

START16 = DevStatus.START << 12
START32 = DevStatus.START << 28


class Device:
    """Base class for all specific PILS devices.

    :param name: The device name.  This is not currently used by Zapf.
    :param addr: The base device address.
    :param typecode: The device's typecode.
    :param info: The "additional info" dictionary returned by the `.Scanner`.
    :param io: A `.PlcIO` instance to use for communication with the PLC.
    :param log: A `logging.Logger` to use for device related logging.
    """

    @classmethod
    def class_for(cls, typecode):
        return TYPECODE_MAP.get(typecode, (None,))[0]

    valuetype = None

    def __init__(self, name, addr, typecode, info, io, log):
        self.name = name
        self.typecode = typecode
        self.device_kind = typecode >> 8
        self.total_size = 2 * (typecode & 0xFF)
        self.info = info
        self.aux_spec = info['aux']
        self.errid_spec = info['errid']
        self._params = info['params']
        self._funcs = info['funcs']

        self._enum_r = info['enum_r']  # int -> str
        self._enum_w = info['enum_w']  # str -> int
        if info['enum_w']:
            self._enum_w_rev = {v: k for (k, v) in self._enum_w.items()}
        else:
            self._enum_w_rev = None

        typeinfo = TYPECODE_MAP[typecode]
        self.value_fmt = typeinfo.value_fmt
        self.value_size = calcsize(self.value_fmt)
        self.status_size = typeinfo.status_size
        self.num_values = typeinfo.num_values
        self.num_params = typeinfo.num_params

        self.status_fmt = 'H'
        self.START = START16
        if self.status_size > 2:
            self.status_fmt = 'I'
            self.START = START32

        self.addr = addr
        next_addr = addr + self.num_values * self.value_size
        if typeinfo.has_target:
            self.target_addr = next_addr
            next_addr += self.num_values * self.value_size
        else:
            self.target_addr = None
        if self.status_size:
            self.status_addr = next_addr
        else:
            self.status_addr = None
        if typeinfo.has_pctrl:
            self.pctrl_addr = self.status_addr + self.status_size
        else:
            self.pctrl_addr = None
        if typeinfo.num_params:
            self.param_addr = self.status_addr + (
                4 if (self.typecode >> 13) == 3 else 2 + self.status_size
            )
        else:
            self.param_addr = None

        if self.device_kind != 5:  # not applicable for MessageIO
            self.value_info = get_valueinfo(self.info, self.value_size)
            # for enums some values may need correction...
            self.value_fmt = self.value_info.fmt
        else:
            self.value_info = None

        self.log = log
        self.io = io
        self._init()

        io.register_cache_range(self.addr, self.total_size)
        self.tables = {}
        for (tname, tinfo) in info['tables'].items():
            self.tables[tname] = Table(io, self, tname, tinfo)

    def read_status(self):
        """Read the status information of the device.

        This is a tuple of ``(state, reason, aux, error_id)``.

        ``state`` is the 4-bit state (see `.DevStatus` for possible values).
        ``reason`` is the 4-bit reason code (see `.ReasonMap`).  ``aux``
        contains the up to 24 AUX bits.  ``error_id`` is a device-defined
        16-bit integer.

        For devices that do not have an error ID field, it is returned as zero.
        """
        if self.status_addr is None:
            if self.target_addr and self.target_addr != self.addr and \
               self.read_value() != self.read_target():
                return (DevStatus.BUSY, 0, 0, 0)
            return (DevStatus.IDLE, 0, 0, 0)
        if self.status_size == 2:
            s = StatusStruct16(self.io.read_u16(self.status_addr))
            return (s.STATE, s.REASON, s.AUX, 0)
        if self.status_size == 4:
            s = StatusStruct32(self.io.read_u32(self.status_addr))
            return (s.STATE, s.REASON, s.AUX, 0)
        if self.status_size == 6:
            value, err_id = self.io.read_fmt(self.status_addr, 'IH')
            s = StatusStruct32(value)
            return (s.STATE, s.REASON, s.AUX, err_id)
        raise SpecError('invalid status_size')

    def decode_aux(self, aux):
        """Decode the aux value returned by `.read_status`.

        Return a string for human consumption.
        """
        return decode_bitfield(aux, self.aux_spec)

    def decode_errid(self, errid):
        """Decode the errid value returned by `.read_status`.

        Return a string for human consumption.
        """
        return decode_bitfield(errid, self.errid_spec)

    def change_status(self, initial_states=(), final_state=0):
        """Change the state of the device.

        Since :ref:`not all state transitions are allowed <pils:state-graph>`,
        you can list the allowed states in *initial_states*.  If that list is
        given and the current state is not contained, ``False`` is returned.
        Else the *final_state* is written and ``True`` is returned.

        If the device does not have a status, ``False`` is always returned.
        """
        if self.status_addr is None:
            return False
        if initial_states:
            state = self.read_status()[0]
            if state not in initial_states:
                return False
        status = StatusStruct16()
        status.STATE = final_state
        if self.status_size == 2:
            self.io.write_u16(self.status_addr, int(status))
        else:
            self.io.write_u32(self.status_addr, int(status) << 16)
        return True

    def reset(self):
        """Reset the device.

        This is normally equivalent to `change_status(..., RESET)` with all
        error or unknown states as the initial states.  However, special
        devices like MessageIO can have a different reset strategy.
        """
        error_stati = set(range(16)) - {
            DevStatus.RESET,
            DevStatus.IDLE,
            DevStatus.WARN,
            DevStatus.START,
            DevStatus.BUSY,
            DevStatus.STOP,
            DevStatus.DIAGNOSTIC_ERROR,
        }
        return self.change_status(error_stati, DevStatus.RESET)

    def list_tables(self):
        """Return a list of the names of the device's tables.

        :ref:`Table <pils:table-descriptor>` objects are accessible as
        `self.tables[name]`.

        If the device supports no tables, an empty list is returned.
        """
        return list(self.tables)

    # to implement:

    def _init(self):
        raise NotImplementedError

    def read_value_raw(self):
        """Read the main value of the device.

        The type of the returned value differs depending on the device type:

        * int for discrete and enum devices
        * float for analog devices
        * list of float for vector devices
        """
        raise NotImplementedError

    def read_value(self):
        """Read the main value like `read_value_raw` but converting enums."""
        value = self.read_value_raw()
        if self._enum_r:
            value = str(self._enum_r.get(value, value))
        return value

    def read_target_raw(self):
        """Read the target of the device.

        The returned value is of the same type as `read_value_raw`.

        If the device is read-only and has no target, `.ApiError` is raised.
        """
        raise ApiError('reading target of a read-only device')

    def read_target(self):
        """Read the target value like `read_target_raw` but converting enums."""
        value = self.read_target_raw()
        if self._enum_w_rev:
            value = str(self._enum_w_rev.get(value, value))
        return value

    def change_target_raw(self, _value):
        """Change the target of the device.

        This will atomically write the new target (whose type must be the same
        as returned by `read_value`) and set the state to `START`.

        If the device is read-only and has no target, `.ApiError` is raised.
        """
        raise ApiError('writing target of a read-only device')

    def change_target(self, value):
        """Change the target, with value checking.

        Like `change_target_raw`, but accepts string values for enum types
        and checks value limits.
        """
        if self._enum_w and isinstance(value, str):
            if value not in self._enum_w:
                raise ApiError(f'{value} is not one of '
                               f'{", ".join(self._enum_w)}')
            value = self._enum_w[value]
        elif (value < self.value_info.min_value or
              value > self.value_info.max_value):
            raise ApiError(f'{value} is not within device limits '
                           f'{self.value_info.min_value}..'
                           f'{self.value_info.max_value}')
        self.change_target_raw(value)

    def list_params(self):
        """Return a list of the names of the device's parameters.

        See :ref:`pils:parameters-functions`.

        If the device supports no parameters, an empty list is returned.
        """
        return []

    def get_param_raw(self, _name):
        """Read the given parameter.

        The value type is int or float, enums are not translated.

        The return value is a tuple of ``(param_cmd, value)`` where
        ``param_cmd`` represents the reply of the parameter state machine,
        see `.ParamCMDs`.

        If the device supports no parameters, `.ApiError` is raised.
        """
        raise ApiError('reading parameter of a device without params')

    def get_param(self, name):
        """Read the given parameter.

        The type of the parameter depends on the parameter, it can be int,
        float, or string (enum).

        The reply from the parameter state machine is translated; if setting
        was not successful, `.ApiError` or `.CommError` is raised.

        If the device supports no parameters, `.ApiError` is raised.
        """
        cmd, value = self.get_param_raw(name)
        if cmd == ParamCMDs.ERR_NO_IDX:
            raise ApiError(f'parameter {name!r} does not exist')
        if cmd == ParamCMDs.ERR_RO:
            raise ApiError(f'parameter {name!r} is read-only')
        if cmd == ParamCMDs.ERR_RETRY:
            raise CommError('parameter interface is busy, retry later')
        value_info = self.get_param_valueinfo(name)
        if value_info.enum_r:
            value = str(value_info.enum_r.get(value, value))
        return value

    def get_param_valueinfo(self, _name):
        """Return the given parameter's type info.

        If the device supports no parameters, `.ApiError` is raised.
        """
        raise ApiError('reading parameter type of a device without params')

    def set_param_raw(self, _name, _value):
        """Set the given parameter.

        The type of *value* must be int or float, enums are not translated.

        The return value is a tuple of ``(param_cmd, value)`` where
        ``param_cmd`` represents the reply of the parameter state machine,
        see `.ParamCMDs`, and ``value`` is the read-back value of the
        parameter (which might differ from the written value due to clamping,
        rounding or other correction the PLC makes).

        If the device supports no parameters, `.ApiError` is raised.
        """
        raise ApiError('writing parameter of a device without params')

    def set_param(self, name, value):
        """Set the given parameter.

        The type of *value* depends on the parameter, it must be int, float, or
        string (enum).  Value limits are checked.

        The reply from the parameter state machine is translated; if setting
        was not successful, `.ApiError` or `.CommError` is raised.  Else,
        the readback value is returned.

        If the device supports no parameters, `.ApiError` is raised.
        """
        value_info = self.get_param_valueinfo(name)
        if value_info.enum_w and isinstance(value, str):
            if value not in value_info.enum_w:
                raise ApiError(f'{value} is not one of {name!r} enum values '
                               f'{", ".join(self._enum_w)}')
            value = value_info.enum_w[value]
        elif (value < value_info.min_value or
              value > value_info.max_value):
            raise ApiError(f'{value} is not within {name!r} limits '
                           f'{value_info.min_value}..'
                           f'{value_info.max_value}')
        cmd, value = self.set_param_raw(name, value)
        if cmd == ParamCMDs.ERR_NO_IDX:
            raise ApiError(f'parameter {name} does not exist')
        if cmd == ParamCMDs.ERR_RO:
            raise ApiError(f'parameter {name} is read-only')
        if cmd == ParamCMDs.ERR_RETRY:
            raise CommError('parameter interface is busy, retry later')
        return value

    def list_funcs(self):
        """Return a list of the names of the device's special functions.

        See :ref:`pils:parameters-functions`.

        If the device supports no special functions, an empty list is returned.
        """
        return []

    def exec_func(self, _name, _value=None):
        """Execute the given special function with an argument.

        The return value is a tuple of ``(param_cmd, value)`` where
        ``param_cmd`` represents the reply of the parameter state machine,
        see `.ParamCMDs`, and ``value`` is the return value of the function.

        If the device supports no special functions, `.ApiError` is raised.
        """
        raise ApiError('executing function of a device without functions')


class DiscreteDevice(Device):
    """Base class used for discrete devices."""

    def _init(self):
        pass

    def read_value_raw(self):
        return self.io.read_fmt(self.addr, self.value_fmt)[0]

    def read_target_raw(self):
        if self.target_addr is None:
            raise ApiError('reading target of a read-only device')
        return self.io.read_fmt(self.target_addr, self.value_fmt)[0]

    def change_target_raw(self, value):
        if self.target_addr is None:
            raise ApiError('writing target of a read-only device')
        if self.status_addr is None:
            self.io.write_fmt(self.target_addr, self.value_fmt, value)
        else:
            self.io.write_fmt(
                self.target_addr,
                self.value_fmt + self.status_fmt,
                value,
                self.START,
            )


class SimpleDiscreteIn(DiscreteDevice):
    """Class for simple discrete input devices.

    See :ref:`pils:dev-simplediscreteinput`,
    :ref:`pils:dev-simplediscrete32input`,
    :ref:`pils:dev-simplediscrete64input`.
    """


class SimpleDiscreteOut(DiscreteDevice):
    """Class for simple discrete output devices.

    See :ref:`pils:dev-simplediscreteoutput`,
    :ref:`pils:dev-simplediscrete32output`,
    :ref:`pils:dev-simplediscrete64output`.
    """


class DiscreteIn(DiscreteDevice):
    """Class for discrete input devices.

    See :ref:`pils:dev-discreteinput`, :ref:`pils:dev-discrete32input`,
    :ref:`pils:dev-discrete64input`.
    """


class DiscreteOut(DiscreteDevice):
    """Class for discrete output devices.

    See :ref:`pils:dev-discreteoutput`, :ref:`pils:dev-discrete32output`,
    :ref:`pils:dev-discrete64output`.
    """


class Keyword(DiscreteDevice):
    """Class for keyword devices.

    See :ref:`pils:dev-keyword`, :ref:`pils:dev-keyword32`,
    :ref:`pils:dev-keyword64`.
    """

    def _init(self):
        self.target_addr = self.addr


class StatusWord(Keyword):
    """Class for :ref:`pils:dev-statusword`, :ref:`pils:dev-extstatusword`."""

    def _init(self):
        super()._init()
        self.status_addr = self.addr


class AnalogDevice(Device):
    """Base class for analog valued devices (except vector devices)."""

    def _init(self):
        pass

    def read_value_raw(self):
        res = self.io.read_fmt(self.addr, self.value_fmt)[0]
        if self.info['basetype'] == 'enum':
            return res & 65535
        return res

    def read_target_raw(self):
        if self.target_addr is None:
            raise ApiError('reading target of a read-only device')
        res = self.io.read_fmt(self.target_addr, self.value_fmt)[0]
        if self.info['basetype'] == 'enum':
            return res & 65535
        return res

    def change_target_raw(self, value):
        if self.target_addr is None:
            raise ApiError('writing target of a read-only device')
        if self.status_addr is None:
            self.io.write_fmt(self.target_addr, self.value_fmt, value)
        else:
            self.io.write_fmt(
                self.target_addr,
                self.value_fmt + self.status_fmt,
                value,
                self.START,
            )


class SimpleAnalogIn(AnalogDevice):
    """Class for simple analog input devices.

    See :ref:`pils:dev-simpleanaloginput`,
    :ref:`pils:dev-simpleanalog64input`.
    """


class SimpleAnalogOut(AnalogDevice):
    """Class for simple analog output devices.

    See :ref:`pils:dev-simpleanalogoutput`,
    :ref:`pils:dev-simpleanalog64output`.
    """


class AnalogIn(AnalogDevice):
    """Class for analog input devices.

    See :ref:`pils:dev-analoginput`, :ref:`pils:dev-analog64input`.
    """


class AnalogOut(AnalogDevice):
    """Class for analog output devices.

    See :ref:`pils:dev-analogoutput`, :ref:`pils:dev-analog64output`.
    """


class RealValue(AnalogDevice):
    """Class for :ref:`pils:dev-realvalue`, :ref:`pils:dev-realvalue64`."""

    def _init(self):
        self.target_addr = self.addr


class FlatParams:
    def _init(self):
        if self.num_params != len(self._params):
            raise SpecError('mismatch between parameter count between typecode '
                            'and parameter indices from indexer '
                            f'({self.num_params}/{len(self._params)})')
        for pn, param in self._params.items():
            if param['idx'] >= self.num_params:
                raise SpecError(f'parameter {pn} has a too big index (must '
                                f'be < {self.num_param})')
        # map parameter name to (addr, typeinfo)
        param_map = {}
        for pn, par in self._params.items():
            idx = par['idx']
            val_info = get_valueinfo(par, self.value_size)
            addr = self.param_addr + idx * self.value_size
            param_map[pn] = (addr, val_info)
            param_map[idx] = (addr, val_info)
        self._param_map = param_map

    def list_params(self):
        return list(self._params)

    def get_param_valueinfo(self, name):
        return self._param_map[name][1]

    def get_param_raw(self, name):
        addr, val_info = self._param_map.get(name, (None,) * 2)
        if addr:
            return ParamCMDs.DONE, self.io.read_fmt(addr, val_info.fmt)[0]
        return ParamCMDs.ERR_NO_IDX, None

    def set_param_raw(self, name, value):
        addr, val_info = self._param_map.get(name, (None,) * 2)
        if not addr:
            return ParamCMDs.ERR_NO_IDX, None
        if val_info.readonly:
            return ParamCMDs.ERR_RO, None
        self.io.write_fmt(addr, val_info.fmt, value or 0)
        return ParamCMDs.DONE, self.io.read_fmt(addr, val_info.fmt)[0]


class FlatIn(FlatParams, AnalogDevice):
    """Class for :ref:`pils:dev-flatinput`, :ref:`pils:dev-flat64input`."""


class FlatOut(FlatParams, AnalogDevice):
    """Class for :ref:`pils:dev-flatoutput`, :ref:`pils:dev-flat64output`."""


class ParamInterface:
    param_timeout = 1

    def _init(self):
        self.param_sm = ParamControl()
        func_map = {}
        param_map = {}
        for pn, func in self._funcs.items():
            idx = func['idx']
            arg_info = res_info = None
            if func['argument']:
                arg_info = get_valueinfo(func['argument'], self.value_size)
            if func['result']:
                res_info = get_valueinfo(func['result'], self.value_size)
            func_map[idx] = (idx, arg_info, res_info)
            func_map[pn] = (idx, func['description'], arg_info, res_info)
        for pn, par in self._params.items():
            idx = par['idx']
            val_info = get_valueinfo(par, self.value_size)
            param_map[idx] = (idx, val_info)
            param_map[pn] = (idx, val_info)
        self._param_map = param_map
        self._func_map = func_map

    def get_param_index(self, name):
        return self._param_map[name][0]

    def get_func_index(self, name):
        return self._func_map[name][0]

    def get_param_valueinfo(self, name):
        return self._param_map[name][1]

    def get_func_valueinfos(self, name):
        return self._func_map[name][1:4]

    def get_pctrl(self, fmt=None, *, cached=True):
        pctrl, self.param_value = self.io.read_pctrl(self.pctrl_addr, fmt,
                                                     cached=cached)
        self.param_sm(pctrl)
        return self.param_sm, self.param_value

    def set_pctrl(self, cmd=None, idx=None, value=None, fmt=None):
        if None in (value, fmt):
            value = fmt = None
        if cmd is None and idx is None:
            pctrl = None
        else:
            if cmd is None or idx is None:
                # aarg! need read-modify-write cycle
                pctrl = self.io.read_pctrl(self.pctrl_addr, cached=False)[0]
                self.param_sm(pctrl)
            if cmd is not None:
                self.param_sm.CMD = cmd
            if idx is not None:
                self.param_sm.IDX = idx
            pctrl = int(self.param_sm)
        self.io.write_pctrl(self.pctrl_addr, fmt, pctrl, value)

    def wait_sm_available(self, fmt=None, *, cached=False):
        timesout = time.monotonic() + self.param_timeout
        while time.monotonic() <= timesout:
            self.get_pctrl(fmt, cached=cached)
            if self.param_sm.available:
                return True
            if int(self.param_sm) == 0:
                # must be because of PLC power cycle
                return True
            # wait for cache update, do not burn CPU cycles...
            if cached:
                self.io.cache.updated.clear()
                self.io.cache.updated.wait(self.io.cache.cycle)
        return False

    def _get_param(self, idx, fmt):
        # XXX? With lock?
        if self.wait_sm_available():
            for _ in range(10):
                self.set_pctrl(ParamCMDs.DO_READ, idx)
                if self.wait_sm_available(fmt) and idx == self.param_sm.IDX:
                    return self.param_sm.CMD, self.param_value
        return ParamCMDs.ERR_RETRY, None

    def _set_param(self, idx, fmt, value):
        # XXX? With lock?
        if self.wait_sm_available():
            for _ in range(10):
                self.set_pctrl(ParamCMDs.DO_WRITE, idx, value, fmt)
                if self.wait_sm_available(fmt) and idx == self.param_sm.IDX:
                    return self.param_sm.CMD, self.param_value
        return ParamCMDs.ERR_RETRY, None

    def get_param_raw(self, name):
        if name not in self._params:
            return ParamCMDs.ERR_NO_IDX, None
        idx, val_info = self._param_map[name]
        return self._get_param(idx, val_info.fmt)

    def set_param_raw(self, name, value):
        if name not in self._params:
            return ParamCMDs.ERR_NO_IDX, None
        idx, val_info = self._param_map[name]
        if val_info.readonly:
            return ParamCMDs.ERR_RO, None
        self._set_param(idx, val_info.fmt, value)
        return self._get_param(idx, fmt=val_info.fmt)

    def list_params(self):
        return list(self._params)

    def list_funcs(self):
        return list(self._funcs)

    def exec_func(self, name, value=None):
        if name not in self._funcs:
            return ParamCMDs.ERR_NO_IDX, None
        idx, _, arg_info, res_info = self._func_map[name]
        # XXX? With lock?
        if self.wait_sm_available():
            if arg_info:
                self.set_pctrl(ParamCMDs.DO_WRITE, idx, value or 0,
                               fmt=arg_info.fmt)
            else:
                self.set_pctrl(ParamCMDs.DO_WRITE, idx)
            # now wait until setting parameter is finished and return
            # read-back-value
            self.wait_sm_available(res_info.fmt if res_info else None)
            return self.param_sm.CMD, self.param_value
        return ParamCMDs.ERR_RETRY, None


class ParamIn(ParamInterface, AnalogDevice):
    """Class for :ref:`pils:dev-paraminput`, :ref:`pils:dev-param64input`."""


class ParamOut(ParamInterface, AnalogDevice):
    """Class for :ref:`pils:dev-paramoutput`, :ref:`pils:dev-param64output`."""


class VectorDevice(Device):
    """Base class for Vector devices."""

    def _init(self):
        pass

    def read_value_raw(self):
        if self.value_size == 4:
            return self.io.read_f32s(self.addr, self.num_values)
        if self.value_size == 8:
            return self.io.read_f64s(self.addr, self.num_values)
        raise SpecError('invalid value_size')

    def read_value(self):
        values = self.read_value_raw()
        if self._enum_r:
            values = [str(self._enum_r.get(value, value)) for value in values]
        return values


class VectorIn(ParamInterface, VectorDevice):
    """Class for vector input devices.

    See :ref:`pils:dev-vectorinput`, :ref:`pils:dev-vector64input`.
    """


class VectorOut(ParamInterface, VectorDevice):
    """Class for vector output devices.

    See :ref:`pils:dev-vectoroutput`, :ref:`pils:dev-vector64output`.
    """

    def read_target_raw(self):
        if self.value_size == 4:
            return self.io.read_f32s(self.target_addr, self.num_values)
        if self.value_size == 8:
            return self.io.read_f64s(self.target_addr, self.num_values)
        raise SpecError('invalid value_size')

    def read_target(self):
        values = self.read_target_raw()
        if self._enum_w_rev:
            values = [str(self._enum_w_rev.get(value, value)) for value in values]
        return values

    def change_target_raw(self, value):
        if self.value_size == 4:
            if self.status_size == 2:
                self.io.write_f32s_u16(self.target_addr, value, START16)
            else:
                self.io.write_f32s_u32(self.target_addr, value, START32)
        elif self.value_size == 8:
            self.io.write_f64s_u32(self.target_addr, value, START32)

    def change_target(self, value):
        if self._enum_w:
            values = []
            for item in value:
                if isinstance(item, str):
                    if item not in self._enum_w:
                        raise ApiError(f'item {item} is not one of '
                                       f'{", ".join(self._enum_w)}')
                    item = self._enum_w[item]  # noqa: PLW2901
                values.append(item)
        else:
            values = value
        self.change_target_raw(values)


class MessageIO(Device):
    """Class for :ref:`pils:dev-messageio`."""

    def _init(self):
        # self.addr is the mailbox register.  Data bytes start at next word.
        self.data_addr = self.addr + 2
        self.max_msg_size = self.total_size - 2
        self.msg_fmt = f'H{self.max_msg_size}s'

    def reset(self):
        # reset by setting the mailbox to initial state
        self.io.write_u16(self.addr, 0)

    def read_value_raw(self):
        raise ApiError('MessageIO has no value')

    def read_target_raw(self):
        raise ApiError('MessageIO has no target')

    def change_target_raw(self, _value):
        raise ApiError('MessageIO has no target')

    def read_status(self):
        """Read the status information of the device.

        This is a tuple of ``(state, reason, aux, error_id)``.

        ``state`` is the 4-bit state (see `.DevStatus` for possible values).
        ``reason`` is the 4-bit reason code (see `.ReasonMap`).  ``aux``
        contains the up to 24 AUX bits.  ``error_id`` is a device-defined
        16-bit integer.

        For devices that do not have an error ID field, it is returned as zero.
        """
        mailbox = self.io.read_u16(self.addr)
        # sensible mapping of reset, idle, busy, busy, busy, busy, busy, error
        return {0: DevStatus.RESET, 1: DevStatus.IDLE, 7: DevStatus.ERROR}.get(
            mailbox >> 13, DevStatus.BUSY), 0, 0, 0

    def _get_mailbox_state(self, mailbox=None):
        """Check mailbox state and raise if that indicates an error."""
        if mailbox is None:
            mailbox = self.io.read_u16(self.addr)
        mailbox_state = mailbox >> 13
        if mailbox_state == 0:
            raise CommError('mailbox in Reset state! is the PLC running?')
        if mailbox_state == 7:
            raise CommError('mailbox in Error state! try reset to clean up')
        return mailbox_state

    def _wait_for(self, state):
        """Wait for a certain state in the mailbox, with timeout."""
        for tri in range(100):
            mailbox_state = self._get_mailbox_state()
            if mailbox_state == state:
                return
            time.sleep(0.010 * tri)  # how to do better?
        mailbox_state = self._get_mailbox_state()
        if mailbox_state == state:
            return
        # is it worth decoding the mailbox states to strings?
        raise CommError(
            f'time-out while waiting for mailbox state {state} '
            f'(stuck at {mailbox_state}?)' # noqa: COM812
        )

    def _wait_for_not(self, state):
        """Wait for a certain state no longer in the mailbox, with timeout."""
        for tri in range(100):
            mailbox_state = self._get_mailbox_state()
            if mailbox_state != state:
                return
            time.sleep(0.010 * tri)  # how to do better?
        mailbox_state = self._get_mailbox_state()
        if mailbox_state != state:
            return
        # is it worth decoding the mailbox states to strings?
        raise CommError('time-out while waiting for mailbox leaving '
                        f'state {state}')

    def communicate(self, request):
        if isinstance(request, bytes):
            # split into approprate sized chunks
            requests = []
            while request:
                part, request = (
                    request[:self.max_msg_size],
                    request[self.max_msg_size:],
                )
                requests.append(part)
            result = self._communicate(requests)
            return b''.join(result)
        return self._communicate(request)

    def _communicate(self, requests):
        """Initiate a request-reply cycle with the PLC.

        Send the given request (a byte array) and return the response.
        """
        for r in requests:
            if len(r) > self.max_msg_size:
                raise ApiError('MessageIO: length of message parts can not '
                               f'exceed {self.max_msg_size} bytes, found '
                               f'{len(r)} bytes')

        # wait for initial state of mailbox
        self._wait_for(1)

        # need to send partial parts first
        while requests:
            r = requests.pop(0)
            if requests:
                # send partial part
                self.io.write_fmt(self.addr, self.msg_fmt, (2 << 13) + len(r), r)
                # wait for ACK
                self._wait_for(3)
            else:
                # send final part
                self.io.write_fmt(self.addr, self.msg_fmt, (4 << 13) + len(r), r)
                # wait until plc reacts
                self._wait_for_not(4)  # should go to 5 or 1 by PLC

        replies = []
        # collect partial replies first
        mailbox, data = self.io.read_fmt(self.addr, self.msg_fmt)
        mailbox_state = self._get_mailbox_state(mailbox)
        replies.append(data[: mailbox & 0x1FF])
        while mailbox_state == 5:
            # ack partial reply
            self.io.write_u16(self.addr, 6 << 13)
            self._wait_for_not(6)
            mailbox, data = self.io.read_fmt(self.addr, self.msg_fmt)
            mailbox_state = self._get_mailbox_state(mailbox)
            replies.append(data[: mailbox & 0x1FF])
        # final part already collected above...
        if mailbox_state == 1:  # last part transferred
            return replies
        raise CommError(f'bad mailbox state! expected 1, got {mailbox_state}')


# Note: has_target means that a target field is there, not that the device
# is considered read-only. For this use readonly....
class Type(NamedTuple):
    devcls: type[Device]
    value_fmt: str
    num_values: int
    has_target: bool
    readonly: bool
    status_size: int
    num_params: int
    has_pctrl: bool


# ruff: noqa: FBT003
TYPECODE_MAP = {
    0x1201: Type(SimpleDiscreteIn,  'h', 1, False, True,  0, 0, False),
    0x1202: Type(SimpleDiscreteIn,  'i', 1, False, True,  0, 0, False),
    0x1204: Type(SimpleDiscreteIn,  'q', 1, False, True,  0, 0, False),
    0x1302: Type(SimpleAnalogIn,    'f', 1, False, True,  0, 0, False),
    0x1304: Type(SimpleAnalogIn,    'd', 1, False, True,  0, 0, False),
    0x1401: Type(Keyword,           'H', 1, False, False, 0, 0, False),
    0x1402: Type(Keyword,           'I', 1, False, False, 0, 0, False),
    0x1404: Type(Keyword,           'Q', 1, False, False, 0, 0, False),
    0x1502: Type(RealValue,         'f', 1, False, False, 0, 0, False),
    0x1504: Type(RealValue,         'd', 1, False, False, 0, 0, False),
    0x1602: Type(SimpleDiscreteOut, 'h', 1, True,  False, 0, 0, False),
    0x1604: Type(SimpleDiscreteOut, 'i', 1, True,  False, 0, 0, False),
    0x1608: Type(SimpleDiscreteOut, 'q', 1, True,  False, 0, 0, False),
    0x1704: Type(SimpleAnalogOut,   'f', 1, True,  False, 0, 0, False),
    0x1708: Type(SimpleAnalogOut,   'd', 1, True,  False, 0, 0, False),
    # changeable via start/Stop/reset/...
    0x1801: Type(StatusWord,        'H', 1, False, True,  2, 0, False),
    # changeable via start/Stop/reset/...
    0x1802: Type(StatusWord,        'I', 1, False, True,  4, 0, False),
    0x1a02: Type(DiscreteIn,        'h', 1, False, True,  2, 0, False),
    0x1a04: Type(DiscreteIn,        'i', 1, False, True,  4, 0, False),
    0x1a08: Type(DiscreteIn,        'q', 1, False, True,  6, 0, False),
    0x1b03: Type(AnalogIn,          'f', 1, False, True,  2, 0, False),
    0x1b04: Type(AnalogIn,          'f', 1, False, True,  4, 0, False),
    0x1b08: Type(AnalogIn,          'd', 1, False, True,  6, 0, False),
    0x1e03: Type(DiscreteOut,       'h', 1, True,  False, 2, 0, False),
    0x1e04: Type(DiscreteOut,       'h', 1, True,  False, 4, 0, False),
    0x1e06: Type(DiscreteOut,       'i', 1, True,  False, 4, 0, False),
    0x1e0c: Type(DiscreteOut,       'q', 1, True,  False, 6, 0, False),
    0x1f05: Type(AnalogOut,         'f', 1, True,  False, 2, 0, False),
    0x1f06: Type(AnalogOut,         'f', 1, True,  False, 4, 0, False),
    0x1f0c: Type(AnalogOut,         'd', 1, True,  False, 6, 0, False),

    0x4006: Type(ParamIn,           'f', 1, False, True,  2, 1, True),
    0x4008: Type(ParamIn,           'f', 1, False, True,  6, 1, True),
    0x400c: Type(ParamIn,           'd', 1, False, True,  6, 1, True),
    0x5008: Type(ParamOut,          'f', 1, True,  False, 2, 1, True),
    0x500a: Type(ParamOut,          'f', 1, True,  False, 6, 1, True),
    0x5010: Type(ParamOut,          'd', 1, True,  False, 6, 1, True),
}

for n in range(16):
    lfin32 =  0x2000 | (n << 8) | ( 6 + 2 * n)
    fin32 =   0x6000 | (n << 8) | ( 6 + 2 * n)
    fin64 =   0x2000 | (n << 8) | (12 + 4 * n)
    lfout32 = 0x3000 | (n << 8) | ( 8 + 2 * n)
    fout32 =  0x7000 | (n << 8) | ( 8 + 2 * n)
    fout64 =  0x3000 | (n << 8) | (16 + 4 * n)
    TYPECODE_MAP[lfin32]  = Type(FlatIn, 'f', 1, False, True, 2, n + 1, False)
    TYPECODE_MAP[fin32]   = Type(FlatIn, 'f', 1, False, True, 4, n + 1, False)
    TYPECODE_MAP[fin64]   = Type(FlatIn, 'd', 1, False, True, 6, n + 1, False)
    TYPECODE_MAP[lfout32] = Type(FlatOut, 'f', 1, True, False, 2, n + 1, False)
    TYPECODE_MAP[fout32]  = Type(FlatOut, 'f', 1, True, False, 4, n + 1, False)
    TYPECODE_MAP[fout64]  = Type(FlatOut, 'd', 1, True, False, 6, n + 1, False)

    if n == 0:
        continue

    vin32 =  0x4000 | (n << 8) | ( 6 + 2 * n)
    vin64 =  0x4000 | (n << 8) | (12 + 4 * n)
    vout32 = 0x5000 | (n << 8) | ( 8 + 4 * n)
    vout64 = 0x5000 | (n << 8) | (16 + 8 * n)
    TYPECODE_MAP[vin32]      = Type(VectorIn,  'f', n + 1, False, True, 2, 1, True)
    TYPECODE_MAP[vin32 + 2]  = Type(VectorIn,  'f', n + 1, False, True, 6, 1, True)
    TYPECODE_MAP[vin64]      = Type(VectorIn,  'd', n + 1, False, True, 6, 1, True)
    TYPECODE_MAP[vout32]     = Type(VectorOut, 'f', n + 1, True, False, 2, 1, True)
    TYPECODE_MAP[vout32 + 2] = Type(VectorOut, 'f', n + 1, True, False, 6, 1, True)
    TYPECODE_MAP[vout64]     = Type(VectorOut, 'd', n + 1, True, False, 6, 1, True)

for n in range(4, 253, 2):
    msgio = 0x0500 | (n >> 1)
    TYPECODE_MAP[msgio] = Type(MessageIO, f'{n-2}s', 1, False, True, 0, 0, False)


def typecode_description(typecode):
    typeinfo = TYPECODE_MAP[typecode]
    name = typeinfo.devcls.__name__
    valuesize = calcsize(typeinfo.value_fmt)
    if typeinfo.num_params > 0 and not typeinfo.has_pctrl:
        name += f'/{typeinfo.num_params}'
    if typeinfo.num_values > 1:
        name += f'/{typeinfo.num_values}'
    if (typecode in (0x1B03, 0x1F05)) or \
       (typeinfo.status_size == 2 and typeinfo.num_params > 0):
        name += ' (32 bit, legacy)'
    elif typecode == 0x1E03:
        name += ' (16 bit, legacy)'
    elif (typecode) >> 8 == 5:
        name += f' ({valuesize} bytes)'
    else:
        name += f' ({valuesize*8} bit)'
    return name

FMT_LIMITS = {
    'h' : (-(2**15), 2**15-1),
    'H' : (0, 2**16-1),
    'i' : (-(2**31), 2**31-1),
    'I' : (0, 2**32-1),
    'q' : (-(2**63), 2**63-1),
    'Q' : (0, 2**64-1),
    'f' : (-3.403823e38, 3.403823e38),
    'd' : (-1.7976931348623157e+308, 1.7976931348623157e+308),
}



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""A simple server to access a simulated PLC via Modbus/TCP or ADS."""

import socketserver
from struct import pack, unpack, unpack_from

from zapf.proto.ads import (
    ADS_ADR_LEN,
    ADS_DEVINFO,
    ADS_READ,
    ADS_WRITE,
    AMS_HEADER_SIZE,
    DEVINFO,
    HEADER,
)


class ModbusError(Exception):
    pass


class ModbusConnectionHandler(socketserver.BaseRequestHandler):

    def handle(self):
        sock = self.request
        while True:
            # read request
            req = sock.recv(8)
            if len(req) != 8:
                return
            tidpid, lgth, unit, func = unpack('>IHBB', req)
            data = sock.recv(lgth - 2)
            try:
                resp = self.handle_req(lgth, func, data)
            except ModbusError as e:
                msg = pack('>IHBBB', tidpid, lgth, unit, func | 0x80,
                           e.args[0])
            except Exception as e:  # noqa: BLE001
                print('during Modbus request: ', e)
                # send as a slave device failure
                msg = pack('>IHBBB', tidpid, lgth, unit, func | 0x80, 4)
            else:
                msg = pack('>IHBB', tidpid, 2 + len(resp), unit, func) + resp
            sock.sendall(msg)

    def handle_req(self, lgth, func, data):
        # decode request
        if len(data) != lgth - 2:      # illegal data value
            raise ModbusError(3)
        addr, = unpack_from('>H', data)
        addr = addr - 0x3000  # map from Beckhoff standard
        if addr < 0 or addr >= 0x1000:
            # illegal data address
            raise ModbusError(2)
        baddr = 2*addr + 0x10000  # map to byte address
        with self.server.cond:
            self.server.cond.wait()  # wait for go ahead for one round
            if func in (3, 4):
                # read data
                nreg, = unpack('>H', data[2:])
                if nreg > 125:
                    raise ModbusError(3)  # illegal data value
                read = self.server.mem.read(baddr, 2*nreg)
                return pack('>B', 2*nreg) + \
                    b''.join(pack('>H', *unpack('H', read[2*i:2*i+2]))
                             for i in range(nreg))
            if func == 6:
                wdata = pack('H', *unpack('>H', data[2:4]))
                self.server.mem.write(baddr, wdata)
                return data
            if func == 16:
                nreg, dbytes = unpack_from('>HB', data[2:])
                assert dbytes == 2*nreg
                wdata = b''.join(pack('H', *unpack('>H', data[2*i+5:2*i+7]))
                                 for i in range(nreg))
                self.server.mem.write(baddr, wdata)
                return data[:4]
            raise ModbusError(1)  # illegal function


class ModbusServer(socketserver.ThreadingTCPServer):
    allow_reuse_address = True
    daemon_threads = True

    def __init__(self, mem, cond, port):
        self.mem = mem
        self.cond = cond
        socketserver.ThreadingTCPServer.__init__(self, ('localhost', port),
                                                 ModbusConnectionHandler)


class AdsError(Exception):
    pass


class AdsConnectionHandler(socketserver.BaseRequestHandler):

    def handle(self):
        sock = self.request
        while True:
            # read header
            req = sock.recv(HEADER.size)
            if len(req) != HEADER.size:
                return
            _, length, dest, source, cmd, flags, data_length, _, invoke_id = \
                HEADER.unpack(req)
            data = sock.recv(data_length)
            try:
                if flags != 4:
                    raise AdsError(0x1C)  # invalid fragment
                if length != 32 + data_length:
                    raise AdsError(0xE)  # invalid length
                ret_data = self.handle_req(cmd, data)
                err = 0
            except AdsError as e:
                ret_data = b''
                err = e.args[0]
            except Exception as e:  # noqa: BLE001
                print('during ADS request: ', e)
                ret_data = b''
                err = 0x1  # internal error
            response = HEADER.pack(
                0, AMS_HEADER_SIZE + 4 + len(ret_data), source, dest, cmd, 5,
                4 + len(ret_data), err, invoke_id) + pack('<I', err) + ret_data
            sock.sendall(response)

    def handle_req(self, cmd, data):
        if cmd == ADS_DEVINFO:
            return DEVINFO.pack(1, 2, 731, b'Zapf sim')
        if cmd == ADS_READ:
            if len(data) != 12:
                raise AdsError(0xE)  # invalid length
            igroup, ioffset, length = ADS_ADR_LEN.unpack(data)
            ioffset += 0x10000
            if igroup != 0x4020:
                raise AdsError(0x702)  # invalid index group
            with self.server.cond:
                self.server.cond.wait()
                return data[8:12] + self.server.mem.read(ioffset, length)
        elif cmd == ADS_WRITE:
            if len(data) < 12:
                raise AdsError(0xE)
            igroup, ioffset, length = ADS_ADR_LEN.unpack(data[:12])
            ioffset += 0x10000
            if igroup != 0x4020:
                raise AdsError(0x702)  # invalid index group
            with self.server.cond:
                self.server.cond.wait()
                self.server.mem.write(ioffset, data[12:])
                return b''
        else:
            raise AdsError(0xB)  # invalid command


class AdsServer(socketserver.ThreadingTCPServer):
    allow_reuse_address = True
    daemon_threads = True

    def __init__(self, mem, cond, port):
        self.mem = mem
        self.cond = cond
        socketserver.ThreadingTCPServer.__init__(self, ('localhost', port),
                                                 AdsConnectionHandler)
